# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the package upload workflow."""

from collections.abc import Sequence
from typing import Any
from unittest.mock import call, patch

import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianUpload,
    DebusinePromise,
    TaskTypes,
)
from debusine.client.models import LookupChildType, RuntimeParameter
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    TaskDatabase,
    WorkRequest,
    WorkflowTemplate,
    default_workspace,
)
from debusine.server.workflows import (
    PackageUploadWorkflow,
    WorkflowValidationError,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    PackageUploadWorkflowData,
    PackageUploadWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import WorkflowTestBase
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    SbuildData,
    SbuildInput,
)


class PackageUploadWorkflowTests(WorkflowTestBase[PackageUploadWorkflow]):
    """Unit tests for :py:class:`PackageUploadWorkflow`."""

    def create_package_upload_workflow(
        self,
        *,
        extra_task_data: dict[str, Any],
        validate: bool = True,
    ) -> PackageUploadWorkflow:
        """Create a package upload workflow."""
        task_data = {
            "target": "sftp://upload.example.org/queue/",
            "require_signature": False,
        }
        task_data.update(extra_task_data)
        wr = self.playground.create_workflow(
            task_name="package_upload", task_data=task_data, validate=validate
        )
        return self.get_workflow(wr)

    def create_source_binary_artifacts(
        self, architectures: Sequence[str] | None = None
    ) -> tuple[Artifact, list[Artifact], Collection]:
        """Return one source and a list of binary packages (one per arch)."""
        if architectures is None:
            architectures = ("amd64",)
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )

        source = self.playground.create_minimal_source_package_artifact()
        uploads = []
        for architecture in architectures:
            upload = self.playground.create_upload_artifacts(
                source=False,
                binary=True,
                binaries=[("hello", architecture)],
            )
            uploads.append(upload.upload)
            for binary in upload.binaries:
                CollectionItem.objects.create_from_artifact(
                    binary,
                    parent_collection=collection,
                    name=f"binary-{binary.data['deb_fields']['Architecture']}",
                    created_by_user=self.playground.get_default_user(),
                    data={},
                    created_by_workflow=None,
                )

        return source, uploads, collection

    def test_validate_input_source_binary_artifacts_not_set(self) -> None:
        """Invalid: source_artifact and binaries_artifacts not set."""
        w = self.create_package_upload_workflow(
            extra_task_data={}, validate=False
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            '"source_artifact" or "binary_artifacts" must be set',
        ):
            w.validate_input()

    def test_validate_input_source_package_vendor_codename_not_set(
        self,
    ) -> None:
        """Raise WorkflowValidationError: vendor and codename required."""
        source = self.playground.create_minimal_source_package_artifact()

        w = self.create_package_upload_workflow(
            extra_task_data={"source_artifact": source.id},
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            '"vendor" and "codename" are required when '
            'source artifact category is debian:source-package',
        ):
            w.validate_input()

    def test_create_orchestrator(self) -> None:
        """A PackageUploadWorkflow can be instantiated."""
        source_artifact, [binary_artifact], _ = (
            self.create_source_binary_artifacts(["arm64"])
        )
        target_distribution = "debian:bookworm"

        w = self.create_package_upload_workflow(
            extra_task_data={
                "source_artifact": source_artifact.id,
                "binary_artifacts": [binary_artifact.id],
                "target_distribution": target_distribution,
                "vendor": "debian",
                "codename": "bookworm",
            }
        )

        self.assertEqual(w.data.source_artifact, source_artifact.id)
        self.assertEqual(
            w.data.binary_artifacts, LookupMultiple((binary_artifact.id,))
        )

        self.assertEqual(w.data.target_distribution, target_distribution)

    def orchestrate(self, *, data: dict[str, Any]) -> WorkRequest:
        """Create a PackageUpload workflow and call orchestrate_workflow."""
        template = WorkflowTemplate.objects.create(
            name="package_upload",
            workspace=default_workspace(),
            task_name="package_upload",
            runtime_parameters=RuntimeParameter.ANY,
        )

        wr = self.playground.create_workflow(
            template,
            task_data={"target": "sftp://upload.example.org/queue", **data},
        )
        self.assertTrue(orchestrate_workflow(wr))

        return wr

    def test_populate_makesourcepackageupload(self) -> None:
        """
        Workflow create two children work request.

        Work requests created:
        -makesourcepackageupload
        -package_upload
        """
        source = self.playground.create_minimal_source_package_artifact()

        with patch.object(
            PackageUploadWorkflow,
            "requires_artifact",
            wraps=PackageUploadWorkflow.requires_artifact,
        ) as requires_artifact:
            wr = self.orchestrate(
                data={
                    "source_artifact": source.id,
                    "target_distribution": "debian:bookworm",
                    "since_version": "1.0",
                    "require_signature": False,
                    "vendor": "debian",
                    "codename": "bullseye",
                }
            )

        self.assertEqual(wr.children.count(), 2)

        [make_source] = wr.children.filter(task_name="makesourcepackageupload")

        self.assertEqual(
            make_source.workflow_data_json,
            {
                "display_name": "Make .changes file",
                "step": "source-package-upload",
            },
        )

        self.assertEqual(
            make_source.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bullseye",
                "input": {"source_artifact": source.id},
                "since_version": "1.0",
                "target_distribution": "debian:bookworm",
            },
        )

        self.assertEqual(make_source.parent, wr)

        self.assert_work_request_event_reactions(
            make_source,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    name_template="package-upload-source",
                )
            ],
        )

        # Assert that self.requires_artifact() was called.
        # It cannot test the dependencies because source_artifact was
        # a real artifact and not a promise (so no dependencies were added)
        self.assertIn(
            call(make_source, source.id), requires_artifact.mock_calls
        )

        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(package_upload.task_type, TaskTypes.SERVER)
        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": (
                    "Package upload internal@collections/"
                    "name:package-upload-source"
                ),
                "step": (
                    "package-upload-internal@collections/"
                    "name:package-upload-source"
                ),
            },
        )
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {
                    "upload": "internal@collections/name:package-upload-source"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        self.assertQuerySetEqual(
            package_upload.dependencies.all(),
            list(
                WorkRequest.objects.filter(task_name="makesourcepackageupload")
            ),
        )

    def test_populate_source_is_package_upload(self) -> None:
        """Category source package is UPLOAD: skip makesourcepackageupload."""
        source = self.playground.create_upload_artifacts().upload

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "target_distribution": "debian:bookworm",
                "since_version": "1.0",
                "require_signature": False,
            }
        )

        self.assertEqual(wr.children.count(), 1)

        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": f"Package upload {source.id}",
                "step": f"package-upload-{source.id}",
            },
        )
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {"upload": source.id},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

    def test_populate_package_uploads_binaries_only(self) -> None:
        """The workflow create children package uploads: binaries only."""
        _, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "binary_artifacts": [binary.id],
                "target_distribution": "debian:bookworm",
                "require_signature": False,
            },
        )

        self.assertEqual(
            wr.children.filter(task_name="packageupload").count(), 1
        )

        # Other properties of the created package is tested on different
        # unit tests

    def test_populate_merge_uploads(self) -> None:
        """Test mergeuploads task is created (merge_uploads=True)."""
        parent = self.playground.create_workflow()
        assert parent.internal_collection is not None

        architectures = ("amd64", "i386")
        source = self.playground.create_minimal_source_package_artifact()
        for architecture in architectures:
            sbuild = parent.create_child_worker(
                task_name="sbuild",
                task_data=SbuildData(
                    input=SbuildInput(source_artifact=source.id),
                    build_architecture=architecture,
                    environment="debian/match:codename=bookworm",
                ),
                workflow_data=WorkRequestWorkflowData(
                    display_name=architecture, step=architecture
                ),
            )
            self.playground.create_bare_data_item(
                parent.internal_collection,
                f"build-{architecture}",
                category=BareDataCategory.PROMISE,
                data=DebusinePromise(
                    promise_work_request_id=sbuild.id,
                    promise_workflow_id=parent.id,
                    promise_category=ArtifactCategory.UPLOAD,
                ),
            )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data={
                "source_artifact": source.id,
                "binary_artifacts": {
                    "collection": "internal@collections",
                    "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                    "name__startswith": "build-",
                },
                "require_signature": False,
                "target": "sftp://upload.example.org/queue",
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "vendor": "debian",
                "codename": "bullseye",
            },
            parent=parent,
        )
        self.assertTrue(orchestrate_workflow(wr))

        self.assertEqual(
            wr.children.filter(task_name="mergeuploads").count(), 1
        )

        [merge_uploads] = wr.children.filter(task_name="mergeuploads")

        self.assertEqual(
            merge_uploads.task_data,
            {
                "input": {
                    "uploads": [
                        "internal@collections/name:package-upload-source"
                    ]
                    + [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in architectures
                    ]
                },
            },
        )

        self.assertQuerySetEqual(
            merge_uploads.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_name__in={"makesourcepackageupload", "sbuild"}
                )
            ),
            ordered=False,
        )

        self.assertEqual(
            wr.children.filter(task_name="packageupload").count(), 1
        )
        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {
                    "upload": (
                        f"internal@collections/name:"
                        f"package-upload-merged-{wr.id}"
                    )
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        # Replace one of the promises with an artifact and run the workflow
        # orchestrator again.  The result is stable: no new MergeUploads
        # work request is created.
        binary, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": architectures[0],
                    "Files": [{"name": f"hello_1.0_{architectures[0]}.deb"}],
                },
            ),
        )
        parent.internal_collection.manager.add_artifact(
            binary,
            user=self.playground.get_default_user(),
            workflow=parent,
            name=f"build-{architectures[0]}",
            replace=True,
        )
        self.assertTrue(orchestrate_workflow(wr))

        self.assertEqual(
            wr.children.filter(task_name="mergeuploads").count(), 1
        )

    def test_upload_debsign(self) -> None:
        """Upload signed artifacts."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "require_signature": False,
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "key": "ABC123",
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [signer] = wr.children.filter(
            task_type=TaskTypes.SIGNING, task_name="debsign"
        )
        self.assertEqual(
            signer.task_data,
            {
                "key": "ABC123",
                "unsigned": (
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                ),
            },
        )
        identifier = f"internal@collections/name:package-upload-merged-{wr.id}"
        self.assertEqual(
            signer.workflow_data_json,
            {
                "display_name": f"Sign upload for {identifier}",
                "step": f"debsign-{identifier}",
            },
        )

        self.assertQuerySetEqual(
            signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
                )
            ),
        )
        signed_name = (
            f"package-upload-signed-"
            f"internal_collections_name_package-upload-merged-{wr.id}"
        )
        self.assert_work_request_event_reactions(
            signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=signed_name,
                    variables=None,
                )
            ],
        )

        [uploader] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        self.assertEqual(
            uploader.workflow_data_json,
            {
                "display_name": (
                    f"Package upload "
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                ),
                "step": (
                    f"package-upload-internal@collections/"
                    f"name:package-upload-merged-{wr.id}"
                ),
            },
        )

    def test_upload_externaldebsign(self) -> None:
        """Upload external signed artifacts."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "require_signature": True,
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [signer] = wr.children.filter(
            task_type=TaskTypes.WAIT, task_name="externaldebsign"
        )
        self.assertEqual(
            signer.task_data,
            {
                "unsigned": (
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                )
            },
        )
        identifier = f"internal@collections/name:package-upload-merged-{wr.id}"
        self.assertEqual(
            signer.workflow_data_json,
            {
                "display_name": (
                    f"Wait for signature on upload for {identifier}"
                ),
                "step": f"external-debsign-{identifier}",
                "needs_input": True,
            },
        )

        self.assertQuerySetEqual(
            signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
                )
            ),
        )
        signed_name = (
            f"package-upload-signed-"
            f"internal_collections_name_package-upload-merged-{wr.id}"
        )
        self.assert_work_request_event_reactions(
            signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=signed_name,
                    variables=None,
                )
            ],
        )

        [uploader] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

    def test_upload_split_signing(self) -> None:
        """Upload artifacts: developer signs source, Debusine signs binary."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "require_signature": True,
                "target_distribution": "debian/match:codename=bookworm",
                "binary_key": "ABC123",
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [source_signer] = wr.children.filter(
            task_type=TaskTypes.WAIT, task_name="externaldebsign"
        )
        source_lookup = "internal@collections/name:package-upload-source"
        self.assertEqual(source_signer.task_data, {"unsigned": source_lookup})
        self.assertEqual(
            source_signer.workflow_data_json,
            {
                "display_name": (
                    f"Wait for signature on upload for {source_lookup}"
                ),
                "step": f"external-debsign-{source_lookup}",
                "needs_input": True,
            },
        )
        self.assertQuerySetEqual(
            source_signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER,
                    task_name="makesourcepackageupload",
                )
            ),
        )
        source_signed_name = (
            "package-upload-signed-"
            "internal_collections_name_package-upload-source"
        )
        self.assert_work_request_event_reactions(
            source_signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    name_template=source_signed_name,
                )
            ],
        )

        [binary_signer] = wr.children.filter(
            task_type=TaskTypes.SIGNING, task_name="debsign"
        )
        binary_lookup = f"{binary.id}@artifacts"
        self.assertEqual(
            binary_signer.task_data,
            {"key": "ABC123", "unsigned": binary_lookup},
        )
        self.assertEqual(
            binary_signer.workflow_data_json,
            {
                "display_name": f"Sign upload for {binary_lookup}",
                "step": f"debsign-{binary_lookup}",
            },
        )
        self.assertQuerySetEqual(
            binary_signer.dependencies.all(), [source_signer]
        )
        binary_signed_name = f"package-upload-signed-{binary.id}_artifacts"
        self.assert_work_request_event_reactions(
            binary_signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=binary_signed_name,
                    variables=None,
                )
            ],
        )

        [source_uploader, binary_uploader] = wr.children.filter(
            task_name="packageupload"
        ).order_by("id")
        self.assertEqual(
            source_uploader.task_data,
            {
                "input": {
                    "upload": f"internal@collections/name:{source_signed_name}"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )
        self.assertEqual(
            source_uploader.workflow_data_json,
            {
                "display_name": f"Package upload {source_lookup}",
                "step": f"package-upload-{source_lookup}",
            },
        )
        self.assertQuerySetEqual(
            source_uploader.dependencies.all(), [source_signer]
        )
        self.assertEqual(
            binary_uploader.task_data,
            {
                "input": {
                    "upload": f"internal@collections/name:{binary_signed_name}"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )
        self.assertEqual(
            binary_uploader.workflow_data_json,
            {
                "display_name": f"Package upload {binary_lookup}",
                "step": f"package-upload-{binary_lookup}",
            },
        )
        self.assertQuerySetEqual(
            binary_uploader.dependencies.all(),
            [binary_signer, source_uploader],
            ordered=False,
        )

    def test_upload_delayed(self) -> None:
        """Upload to a delayed queue."""
        source = self.playground.create_minimal_source_package_artifact()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "delayed_days": 3,
                "vendor": "debian",
                "codename": "bookworm",
            }
        )

        signed_name = (
            "package-upload-signed-"
            "internal_collections_name_package-upload-source"
        )
        uploader = wr.children.filter(task_name="packageupload").get()
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": 3,
            },
        )

    def test_orchestrate_idempotent(self) -> None:
        """Calling orchestrate twice does not create new work requests."""
        self.playground.create_collection(
            "test", CollectionCategory.TASK_CONFIGURATION
        )
        source, _, binaries_collection = self.create_source_binary_artifacts(
            architectures=("amd64", "i386")
        )

        # Force a predictable default ordering.
        with patch(
            "debusine.db.models.collections.CollectionItem._meta.ordering",
            ["id"],
        ):
            wr = self.orchestrate(
                data={
                    "task_configuration": (
                        f"test@{CollectionCategory.TASK_CONFIGURATION}"
                    ),
                    "source_artifact": source.id,
                    "binary_artifacts": {
                        "collection": binaries_collection.id,
                        "name__startswith": "binary-",
                    },
                    "vendor": "debian",
                    "codename": "bookworm",
                    "merge_uploads": True,
                },
            )

        children = set(wr.children.all())

        # Force a different predictable default ordering.
        with patch(
            "debusine.db.models.collections.CollectionItem._meta.ordering",
            ["-id"],
        ):
            self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), children, ordered=False)

    def test_compute_system_required_tags(self) -> None:
        source_artifact = (
            self.playground.create_minimal_source_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                source_artifact=source_artifact.id,
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
        )
        workflow = self.get_workflow(wr)

        self.assertCountEqual(
            workflow.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    def test_compute_dynamic_data_source_artifact(self) -> None:
        source_artifact = (
            self.playground.create_minimal_source_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                source_artifact=source_artifact.id,
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
        )
        workflow = self.get_workflow(wr)

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(wr)),
            PackageUploadWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → sftp://example.org",
                source_artifact_id=source_artifact.id,
                binary_artifacts_ids=[],
            ),
        )

    def test_compute_dynamic_data_source_is_promise(self) -> None:
        parent = self.playground.create_workflow()
        assert parent.internal_collection is not None

        signing = parent.create_child_worker(
            task_name="noop",
        )
        self.playground.create_bare_data_item(
            parent.internal_collection,
            "source",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=signing.id,
                promise_workflow_id=parent.id,
                promise_category=ArtifactCategory.SOURCE_PACKAGE,
                **{"source_package_name": "hello"},  # noqa: PIE804
            ),
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                source_artifact="internal@collections/name:source",
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
            parent=parent,
        )

        workflow = self.get_workflow(wr)
        self.assertEqual(
            workflow.build_dynamic_data(TaskDatabase(wr)),
            PackageUploadWorkflowDynamicData(
                parameter_summary="hello → sftp://example.org",
                subject="hello",
                binary_artifacts_ids=[],
            ),
        )

    def test_compute_dynamic_data_binary_artifacts(self) -> None:
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    [binary_artifact.id]
                ),
                source_artifact=None,
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
        )
        workflow = self.get_workflow(wr)

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(wr)),
            PackageUploadWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → sftp://example.org",
                source_artifact_id=None,
                binary_artifacts_ids=[binary_artifact.id],
            ),
        )

    def test_get_input_artifacts_source_artifact(self) -> None:
        source_artifact = (
            self.playground.create_minimal_source_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                source_artifact=source_artifact.id,
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
        )
        workflow = self.get_workflow(wr)

        workflow.dynamic_data = workflow.build_dynamic_data(
            TaskDatabase(workflow.work_request)
        )

        assert workflow.data.source_artifact is not None

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=source_artifact.id,
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[],
                ),
            ],
        )

    def test_get_input_artifacts_binary_artifacts(self) -> None:
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data=PackageUploadWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    [binary_artifact.id]
                ),
                source_artifact=None,
                target="sftp://example.org",
                vendor="debian",
                codename="trixie",
            ),
        )
        workflow = self.get_workflow(wr)

        workflow.dynamic_data = workflow.build_dynamic_data(
            TaskDatabase(workflow.work_request)
        )

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[binary_artifact.id],
                )
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Generate cloud-init scripts for bootstrapping Debusine Workers."""

from base64 import b64encode
from textwrap import dedent

import yaml
from django.conf import settings

from debusine.db.models import Token
from debusine.server.worker_pools.models import (
    BaseWorkerPoolDataModel,
    DebianRelease,
    DebusineInstallSource,
)

DAILY_BUILD_SOURCE = dedent(
    """\
    Types: deb
    URIs: https://freexian-team.pages.debian.net/debusine/repository/
    Suites: trixie
    Components: main
    Architectures: all
    Signed-By:
      -----BEGIN PGP PUBLIC KEY BLOCK-----
      .
      mQENBGUAlgcBCAD0AhJt2KVLlrJ+Bx8VcjxCNGcOEMPOY1Db+F+8TiaYrSqZ07kx
      VOJYATlNAqCkQqCu89E9WvqKTg2kyNZchyArwi4bUhEfA+BT+KBeb4dGcoAdwUNX
      c4hx94M/Ow2hAStIni2TUYau99e5sY2QIFb03Lb3dShlQA2EesjQcvnpZCDBcVHW
      kYGj8zjh+i0QdtGVYJlXI/IskSJKexJuiOQ9uIGtsJ+VAm2hD1dTDWvGoUp2quWP
      uHGBouyzFkFQpCOXKBnfpF4h7LMFJSx9KNV/rSMZgLf0F9ukeGnjsST3okQGz2Sz
      E5WdoJuaLOgj/eCeTG2lHD1sdzyndQtG9sKJABEBAAG0Q0RlYnVzaW5lIFNhbHNh
      IFJlcG9zaXRvcnkgPHJhcGhhZWwrZGVidXNpbmUrc2Fsc2FyZXBvQGZyZWV4aWFu
      LmNvbT6JAU4EEwEKADgWIQSQH9XztPO3VNW3w993YWGYlEiOAAUCZQCWBwIbLwUL
      CQgHAwUVCgkICwUWAgMBAAIeAQIXgAAKCRB3YWGYlEiOAO8eCACwcqdDbVigGiv8
      x9iUOp59/6XMG/mtdSjzkkK5LVUEq9bkQAdLRbzb43p0vLl7E/7Yt3on62D6N3QB
      1ap710QsDRIN/CoKRi6sjHCiAqpSUC9LcwgLmnrLwUHDPDrkpsnWmQ7YINtV1P6E
      BFgdqqtw/1avMyliBbh3/XNRQkYq6xkcEZGIftqwX8GYB3M+cheefYe24H/JwLJm
      gerMz8MVEGKMzDm0OaLjWixV6Ga4Wkzo5ya1zg5OmROULAj3CsKDN76OCEwGGMBs
      kNsTB9aAisP8D6wW6ZXX5eChRLRCmKUEB5zpHJ++jdJ0yY38kF+N17w5eos+rc4s
      bMld2SQP
      =XfKL
      -----END PGP PUBLIC KEY BLOCK-----
    """
)

FIREWALL = dedent(
    """\
    #!/usr/sbin/nft -f

    add table inet filter;
    flush table inet filter;

    table inet filter {
        chain input {
            type filter hook input priority filter;
            iif "lo" accept;  # codespell:ignore iif
            tcp dport ssh counter accept;
            udp dport dhcpv6-client counter accept;
            meta l4proto {icmp, icmpv6} counter accept;
            iifname "debusinebr0" udp dport {domain, bootps} accept;
            ct state invalid,new,untracked counter drop;
        }
        chain forward {
            type filter hook forward priority filter;
            # Cloud Metadata endpoints
            ip daddr 169.254.169.254 counter drop;
            ip6 daddr fd00:ec2::254 counter drop;
        }
        chain output {
            type filter hook output priority filter;
            # Cloud Metadata endpoints
            ip daddr 169.254.169.254 meta skuid != 0 counter drop;
            ip6 daddr fd00:ec2::254 meta skuid != 0 counter drop;
        }
    }
    """
)


class AptConfig(BaseWorkerPoolDataModel):
    """Model for cloud-init APT Configure object."""

    preserve_sources_list: bool
    sources_list: str


class WriteFile(BaseWorkerPoolDataModel):
    """Model for cloud-init Write Files write_files objects."""

    path: str
    content: str
    permissions: str | None = None
    owner: str | None = None


class FilesystemSetup(BaseWorkerPoolDataModel):
    """Model for cloud-init Disk Setup fs_setup objects."""

    device: str
    filesystem: str


class CloudInitConfig(BaseWorkerPoolDataModel):
    """Model for cloud-init configuration."""

    apt: AptConfig | None = None
    device_aliases: dict[str, str] = {}
    fs_setup: list[FilesystemSetup] = []
    hostname: str | None = None
    mounts: list[list[str]] = []
    package_reboot_if_required: bool = False
    package_update: bool = False
    package_upgrade: bool = False
    packages: list[str] = []
    bootcmd: list[str] = []
    runcmd: list[str] = []
    write_files: list[WriteFile] = []

    def as_str(self) -> str:
        """Render configuration file in YAML."""
        return "#cloud-config\n" + yaml.safe_dump(
            self.model_dump(exclude_defaults=True)
        )

    def as_base64_str(self) -> str:
        """Render configuration file in YAML and base64-encode the result."""
        return b64encode(self.as_str().encode("utf-8")).decode("ascii")


def worker_bootstrap_cloud_init(
    worker_name: str,
    activation_token: Token,
    debusine_install_source: DebusineInstallSource,
    debian_release: DebianRelease | None,
) -> CloudInitConfig:
    """Generate cloud-init configuration for a new instance."""
    if getattr(activation_token, "key", None) is None:
        raise ValueError(
            f"Requires a fresh token with a key. Token<{activation_token.pk}> "
            f"does not have one."
        )

    if debusine_install_source == DebusineInstallSource.PRE_INSTALLED:
        return CloudInitConfig(
            package_update=True,
            package_upgrade=True,
            package_reboot_if_required=True,
            hostname=worker_name,
            write_files=[
                WriteFile(
                    path="/etc/debusine/worker/activation-token",
                    permissions="0o600",
                    content=activation_token.key,
                    owner="debusine-worker:debusine-worker",
                ),
            ],
            runcmd=["systemctl restart debusine-worker.service"],
        )

    config = CloudInitConfig(
        package_update=True,
        package_upgrade=True,
        package_reboot_if_required=True,
        apt=None,
        bootcmd=[
            # Executes early in boot, before apt is configured
            (
                "[ -e /usr/bin/eatmydata ] "
                "|| (apt-get update && apt-get -y install eatmydata)"
            ),
            # Don't start the worker until we're ready
            "systemctl mask debusine-worker.service",
        ],
        hostname=worker_name,
        write_files=[
            # We set the owner in a runcmd after the user has been created
            WriteFile(
                path="/etc/debusine/worker/activation-token",
                permissions="0o600",
                content=activation_token.key,
            ),
            WriteFile(
                path="/etc/nftables.conf",
                permissions="0o755",
                content=FIREWALL,
            ),
            WriteFile(
                path="/etc/debusine/worker/config.ini",
                content=dedent(
                    f"""\
                    [General]
                    api-url = https://{settings.DEBUSINE_FQDN}/api
                    log-file = /var/log/debusine/worker/worker.log
                    log-level = DEBUG
                    """
                ),
            ),
        ],
        packages=[
            "autopkgtest",
            "ca-certificates",
            "debusine-worker",
            "gpg",
            "incus",
            "mmdebstrap",
            "nftables",
            "sbuild",
            "uidmap",
        ],
        runcmd=[
            (
                "chown debusine-worker:debusine-worker "
                "/etc/debusine/worker/activation-token"
            ),
            "/usr/share/doc/debusine-worker/examples/configure-worker-incus.sh",
            "systemctl enable nftables.service",
            "systemctl start nftables.service",
            "systemctl unmask debusine-worker.service",
            "systemctl restart debusine-worker.service",
        ],
    )
    backports_packages: list[str] = []
    if debusine_install_source == DebusineInstallSource.BACKPORTS:
        backports_packages.append("src:debusine")

    if backports_packages:
        if debian_release:
            pin_release = f"n={debian_release}"
        else:
            pin_release = "a=stable"
        config.write_files += [
            WriteFile(
                path="/etc/apt/preferences.d/90-debusine-backports",
                content=dedent(
                    f"""\
                    Package: {" ".join(backports_packages)}
                    Pin: release {pin_release}-backports
                    Pin-Priority: 900

                    Package: {" ".join(backports_packages)}
                    Pin: release {pin_release}
                    Pin-Priority: -1
                    """
                ),
            ),
        ]
    if debusine_install_source == DebusineInstallSource.DAILY_BUILDS:
        config.write_files += [
            WriteFile(
                path="/etc/apt/sources.list.d/debusine.sources",
                content=DAILY_BUILD_SOURCE,
            ),
            WriteFile(
                path="/etc/apt/preferences.d/90-debusine-ignore-debian",
                content=dedent(
                    """\
                    Package: src:debusine
                    Pin: release o=Debian Backports
                    Pin-Priority: -1

                    Package: src:debusine
                    Pin: release o=Debian
                    Pin-Priority: -1
                    """
                ),
            ),
        ]
    return config



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Server-side task interface."""

import inspect
import logging
import tempfile
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    ClassVar,
    IO,
    Literal,
    TYPE_CHECKING,
    TextIO,
    Union,
    overload,
    override,
)

from debusine.artifacts import WorkRequestDebugLogs
from debusine.artifacts.models import (
    CollectionCategory,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    WorkRequest,
    Worker,
    Workspace,
)
from debusine.tasks import BaseTask, TaskConfigError
from debusine.tasks.models import (
    ActionRecordInTaskHistory,
    BaseDynamicTaskData,
    BaseTaskData,
    EventReaction,
    WorkerType,
)

# TODO: remove
from debusine.tasks.server import TaskDatabaseInterface
from debusine.utils import extract_generic_type_arguments
from debusine.worker.system_information import native_architecture

if TYPE_CHECKING:
    from _typeshed import OpenBinaryModeWriting, OpenTextModeWriting


class DBTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](metaclass=ABCMeta):
    """
    Base class for server-side task-specific logic.

    A DBTask object supports Debusine in overseeing the task execution, with
    operations like applying task configuration, resolving its task data and
    computing scheduler tags.

    This gives the server a single API to access logic that is specific to all
    supported types of tasks, whether they are run server-side or worker-side.
    """

    #: Class used as the in-memory representation of task data.
    task_data_type: type[TD]
    data: TD

    #: Class used as the in-memory representation of dynamic task data.
    dynamic_task_data_type: type[DTD]
    dynamic_data: DTD | None

    #: Must be overridden by child classes to document the current version of
    #: the task's code. A task will only be scheduled on a worker if its task
    #: version is the same as the one running on the scheduler.
    TASK_VERSION: int | None = None

    #: The worker type must be suitable for the task type.  TaskTypes.WORKER
    #: requires an external worker; TaskTypes.SERVER requires a Celery
    #: worker; TaskTypes.SIGNING requires a signing worker.
    TASK_TYPE: TaskTypes

    name: ClassVar[str]
    _sub_tasks: dict[TaskTypes, dict[str, type["DBTask['Any', 'Any']"]]] = (
        defaultdict(dict)
    )

    #: Work request backing this task in the DB
    work_request: WorkRequest

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Register the subclass into DBTask._sub_tasks.

        Used by DBTask.class_from_name() to return the class given the name.
        """
        super().__init_subclass__(**kwargs)

        # The name of the task. It is computed by converting the class name
        # to lowercase.
        cls.name = getattr(cls, "TASK_NAME", cls.__name__.lower())

        # The task data types, computed by introspecting the type arguments
        # used to specialize this generic class.
        [
            cls.task_data_type,
            cls.dynamic_task_data_type,
        ] = extract_generic_type_arguments(cls, DBTask)

        if inspect.isabstract(cls):
            # Don't list abstract base classes as tasks.
            return

        registry = cls._sub_tasks[cls.TASK_TYPE]

        # The same sub-task could register twice
        # (but assert that is the *same* class, not a different
        # subtask with a name with a different capitalisation)
        if cls.name in registry and registry[cls.name] != cls:
            raise AssertionError(f'Two Tasks with the same name: {cls.name!r}')

        # Make sure SERVER and WORKER do not have conflicting task names
        match cls.TASK_TYPE:
            case TaskTypes.SERVER:
                if cls.name in cls._sub_tasks[TaskTypes.WORKER]:
                    raise AssertionError(
                        f'{cls.name!r} already registered as a Worker task'
                    )
            case TaskTypes.WORKER:
                if cls.name in cls._sub_tasks[TaskTypes.SERVER]:
                    raise AssertionError(
                        f'{cls.name!r} already registered as a Server task'
                    )

        registry[cls.name] = cls

    def __init__(self, work_request: WorkRequest) -> None:
        """Initialize the task."""
        self.work_request = work_request

        #: A :py:class:`logging.Logger` instance that can be used in child
        #: classes when you override methods to implement the task.
        self.logger = logging.getLogger("debusine.tasks")

        #: Validate task data and dynamic data
        try:
            self.data = self.task_data_type(**work_request.used_task_data)
            self.dynamic_data = (
                None
                if work_request.dynamic_task_data is None
                else self.dynamic_task_data_type(
                    **work_request.dynamic_task_data
                )
            )
        except ValueError as exc:
            raise TaskConfigError(None, original_exception=exc)

        # Task is aborted: the task does not need to be executed, and can be
        # stopped if it is already running
        self._aborted = False

        # fetch_input() add the downloaded artifacts. Used by
        # `DBTask._upload_work_request_debug_logs()` and maybe by
        # required method `upload_artifacts()`.
        #
        # This is distinct from get_input_artifacts_ids, which is used to
        # extract IDs from dynamic_data for use by UI views
        self._source_artifacts_ids: list[int] = []

        self._debug_log_files_directory: (
            None | (tempfile.TemporaryDirectory[str])
        ) = None

        self.post_init()

    @property
    def workspace(self) -> Workspace:
        """
        Return the workspace for this task.

        This is a shortcut for ``self.work_request.workspace``.
        """
        return self.work_request.workspace

    def set_worker(self, worker: Worker) -> None:
        """Set the worker for this task."""
        # Nothing to do by default

    def post_init(self) -> None:
        """Specific post-init code."""

    def append_to_log_file(self, filename: str, lines: list[str]) -> None:
        """
        Open log file and write contents into it.

        :param filename: use self.open_debug_log_file(filename)
        :param lines: write contents to the logfile
        """
        with self.open_debug_log_file(filename) as file:
            file.writelines([line + "\n" for line in lines])

    @classmethod
    def prefix_with_task_name(cls, text: str) -> str:
        """:return: the ``text`` prefixed with the task name and a colon."""
        if cls.TASK_TYPE is TaskTypes.WORKER:
            # Worker tasks are left unprefixed for compatibility
            return f"{cls.name}:{text}"
        else:
            return f"{cls.TASK_TYPE.lower()}:{cls.name}:{text}"

    def build_architecture(self) -> str | None:
        """
        Return the architecture to run this task on.

        Tasks where build_architecture is not determined by
        self.data.build_architecture should re-implement this method.
        """
        return native_architecture()

    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """
        Check if the specified worker can run the task.

        This method shall take its decision solely based on the supplied
        ``worker_metadata`` and on the configured task data (``self.data``).

        The default implementation always returns True unless
        :py:attr:`TASK_TYPE` doesn't match the worker type or there's a
        mismatch between the :py:attr:`TASK_VERSION` on the scheduler side
        and on the worker side.

        Derived objects can implement further checks by overriding the method
        in the following way::

            if not super().can_run_on(worker_metadata):
                return False

            if ...:
                return False

            return True

        :param dict worker_metadata: The metadata collected from the worker by
            running :py:meth:`analyze_worker` on all the tasks on the worker
            under consideration.
        :return: the boolean result of the check.
        :rtype: bool.
        """
        worker_type = worker_metadata.get("system:worker_type")
        if (self.TASK_TYPE, worker_type) not in {
            (TaskTypes.WORKER, WorkerType.EXTERNAL),
            (TaskTypes.SERVER, WorkerType.CELERY),
            (TaskTypes.SIGNING, WorkerType.SIGNING),
        }:
            return False

        version_key_name = self.prefix_with_task_name("version")
        if worker_metadata.get(version_key_name) != self.TASK_VERSION:
            return False

        # Some tasks might not have "build_architecture"
        task_architecture = self.build_architecture()
        if (
            task_architecture is not None
            and task_architecture
            not in worker_metadata.get("system:architectures", [])
        ):
            return False

        return True

    def compute_system_required_tags(self) -> set[str]:
        """Compute the system set of task-required tags."""
        # Prevent circular import
        import debusine.worker.tags as wtags

        required_tags: set[str] = set()

        # The worker must have code that can handle this task version.
        task_version_tag = (
            wtags.TASK_PREFIX + f"{self.TASK_TYPE.lower()}:{self.name}:"
            f"version:{self.TASK_VERSION}"
        )

        # The worker must be of a type that can run this task.
        match self.TASK_TYPE:
            case TaskTypes.WORKER:
                required_tags.add(wtags.WORKER_TYPE_EXTERNAL)
                required_tags.add(task_version_tag)
            case TaskTypes.SERVER:
                required_tags.add(wtags.WORKER_TYPE_CELERY)
                required_tags.add(task_version_tag)
            case TaskTypes.SIGNING:
                required_tags.add(wtags.WORKER_TYPE_SIGNING)
                required_tags.add(task_version_tag)
            case _:
                # Other tasks should never be dispatched to workers, but
                # make sure of that by requiring a tag that will never
                # exist.
                required_tags.add(wtags.WORKER_TYPE_NOT_ASSIGNABLE)

        return required_tags

    @abstractmethod
    def build_dynamic_data(self, task_database: TaskDatabaseInterface) -> DTD:
        """
        Build a dynamic task data structure for this task.

        :param task_database: TaskDatabaseInterface to use for lookups
        :returns: the newly created dynamic task data
        """

    def compute_dynamic_data(self, task_database: TaskDatabaseInterface) -> DTD:
        """
        Compute dynamic data for this task.

        This may involve resolving artifact lookups.
        """
        return self.build_dynamic_data(task_database)

    def execute_logging_exceptions(self) -> WorkRequestResults:
        """Execute self.execute() logging any raised exceptions."""
        try:
            return self.execute()
        except Exception as exc:
            self.logger.info("Exception in Task %s", self.name, exc_info=True)
            raise exc

    def execute(self) -> WorkRequestResults:
        """
        Call the _execute() method, upload debug artifacts.

        See _execute() for more information.

        :return: result of the _execute() method.
        """  # noqa: D402
        result = self._execute()
        self._upload_work_request_debug_logs()
        return result

    @abstractmethod
    def _execute(self) -> WorkRequestResults:
        """
        Execute the requested task.

        The task must first have been configured. It is allowed to take
        as much time as required. This method will only be run on a worker. It
        is thus allowed to access resources local to the worker.

        It is recommended to fail early by raising a :py:exc:TaskConfigError if
        the parameters of the task let you anticipate that it has no chance of
        completing successfully.

        :return: SUCCESS to indicate success, FAILURE for a failure, ERROR for
            an internal error, SKIPPED if the task turned out to be a noop.
        :raises TaskConfigError: if the parameters of the work request are
            incompatible with the worker.
        """

    def abort(self) -> None:
        """Task does not need to be executed. Once aborted cannot be changed."""
        self._aborted = True

    @property
    def aborted(self) -> bool:
        """
        Return if the task is aborted.

        Tasks cannot transition from aborted -> not-aborted.
        """
        return self._aborted

    @staticmethod
    def class_from_name(
        task_type: TaskTypes, task_name: str
    ) -> type["DBTask['Any', 'Any']"]:
        """
        Return class for :param task_name (case-insensitive).

        :param task_type: type of task to look up

        __init_subclass__() registers DBTask subclasses' into
        DBTask._sub_tasks.
        """
        if (registry := DBTask._sub_tasks.get(task_type)) is None:
            raise ValueError(f"{task_type!r} is not a registered task type")

        task_name_lowercase = task_name.lower()
        if (cls := registry.get(task_name_lowercase)) is None:
            raise ValueError(
                f"{task_name_lowercase!r} is not a registered"
                f" {task_type} task_name"
            )

        return cls

    @staticmethod
    def is_valid_task_name(task_type: TaskTypes, task_name: str) -> bool:
        """Return True if task_name is registered (its class is imported)."""
        if (registry := DBTask._sub_tasks.get(task_type)) is None:
            return False
        return task_name.lower() in registry

    @staticmethod
    def task_names(task_type: TaskTypes) -> list[str]:
        """Return list of sub-task names."""
        return sorted(DBTask._sub_tasks[task_type])

    @staticmethod
    def is_worker_task(task_name: str) -> bool:
        """Check if task_name is a task that can run on external workers."""
        return task_name.lower() in DBTask._sub_tasks[TaskTypes.WORKER]

    @staticmethod
    def worker_task_names() -> list[str]:
        """Return list of sub-task names not of type TaskTypes.SERVER."""
        return sorted(DBTask._sub_tasks[TaskTypes.WORKER].keys())

    def get_input_artifacts_ids(self) -> list[int]:
        """
        Return the list of input artifact IDs used by this task.

        This refers to the artifacts actually used by the task. If
        dynamic_data is empty, this returns the empty list.

        This is used by views to show what artifacts were used by a task.
        `_source_artifacts_ids` cannot be used for this purpose because it is
        only set during task execution.
        """
        if self.dynamic_data is None:
            return []
        return self.dynamic_data.get_input_artifacts_ids()

    def get_label(self) -> str:
        """Return a short human-readable label for the task."""
        return self.work_request.get_label()

    def get_event_reactions(
        self,
        event_name: Literal[
            "on_creation",
            "on_unblock",
            "on_assignment",
            "on_success",
            "on_failure",
        ],
    ) -> list[EventReaction]:
        """
        Return event reactions for this task.

        This allows tasks to provide actions that are processed by the
        server at various points in the lifecycle of the work request.
        """
        event_reactions: list[EventReaction] = []
        if event_name in {"on_success", "on_failure"}:
            event_reactions.append(ActionRecordInTaskHistory())
        return event_reactions

    @overload
    def open_debug_log_file(
        self, filename: str, *, mode: "OpenTextModeWriting" = "a"
    ) -> TextIO: ...

    @overload
    def open_debug_log_file(
        self, filename: str, *, mode: "OpenBinaryModeWriting"
    ) -> BinaryIO: ...

    def open_debug_log_file(
        self,
        filename: str,
        *,
        mode: Union["OpenTextModeWriting", "OpenBinaryModeWriting"] = "a",
    ) -> IO[Any]:
        """
        Open a temporary file and return it.

        The files are always for the same temporary directory, calling it twice
        with the same file name will open the same file.

        The caller must call .close() when finished writing.
        """
        if self._debug_log_files_directory is None:
            self._debug_log_files_directory = tempfile.TemporaryDirectory(
                prefix="debusine-task-debug-log-files-"
            )

        debug_file = Path(self._debug_log_files_directory.name) / filename
        return debug_file.open(mode)

    def _upload_work_request_debug_logs(self) -> None:
        """
        Create a WorkRequestDebugLogs artifact and upload the logs.

        The logs might exist in self._debug_log_files_directory and were
        added via self.open_debug_log_file() or self.create_debug_log_file().

        For each self._source_artifacts_ids: create a relation from
        WorkRequestDebugLogs to source_artifact_id.
        """
        if self._debug_log_files_directory is None:
            return

        work_request_debug_logs = WorkRequestDebugLogs.create(
            files=Path(self._debug_log_files_directory.name).glob("*")
        )

        artifact = Artifact.objects.create_from_local_artifact(
            work_request_debug_logs,
            self.work_request.workspace,
            created_by_work_request=self.work_request,
        )

        for source_artifact_id in self._source_artifacts_ids:
            ArtifactRelation.objects.create(
                artifact=artifact,
                target_id=source_artifact_id,
                type=ArtifactRelation.Relations.RELATES_TO,
            )

        self._debug_log_files_directory.cleanup()
        self._debug_log_files_directory = None

    def compute_system_provided_tags(self) -> set[str]:
        """Compute the system set of task-provided tags."""
        # Prevent circular import
        import debusine.tasks.tags as ttags

        ws = self.work_request.workspace

        provided_tags: set[str] = set()
        provided_tags.add(ttags.SCOPE_PREFIX + ws.scope.name)
        provided_tags.add(ttags.WORKSPACE_PREFIX + f"{ws.scope.name}:{ws.name}")

        for group in self.work_request.created_by.debusine_groups.values(
            "scope__name", "workspace__name", "name"
        ):
            provided_tags.add(
                ttags.GROUP_PREFIX
                + group["scope__name"]
                + ":"
                + (group["workspace__name"] or "")
                + ":"
                + group["name"]
            )

        return provided_tags

    def compute_user_provided_tags(
        self, dynamic_data: BaseDynamicTaskData
    ) -> set[str]:
        """Compute the user set of task-provided tags."""
        # Prevent circular import
        import debusine.tasks.tags as ttags

        provided_tags: set[str] = set()
        if package_name := dynamic_data.get_source_package_name():
            provided_tags.add(ttags.SOURCE_PACKAGE_PREFIX + package_name)
        return provided_tags

    def apply_task_configuration(self) -> None:
        """Apply task configuration to this task and work request."""
        import debusine.server.tags as tags
        import debusine.tasks.tags as ttags
        import debusine.worker.tags as wtags
        from debusine.db.models.task_database import TaskDatabase
        from debusine.server.collections.debusine_task_configuration import (
            apply_configuration,
        )

        self.work_request.configured_task_data = None
        self.work_request.dynamic_task_data = None

        # Lookup the task-configuration collection to use
        config_collection: Collection | None = None
        if (task_configuration := self.data.task_configuration) is not None:
            # Look up the debusine:task-configuration collection
            try:
                lookup_result = self.work_request.lookup_single(
                    lookup=task_configuration,
                    default_category=CollectionCategory.TASK_CONFIGURATION,
                    expect_type=LookupChildType.COLLECTION,
                )
            except KeyError:
                if (
                    task_configuration
                    != BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP
                ):
                    raise
            else:
                config_collection = lookup_result.collection

        # Compute the configured task data
        self.work_request.configured_task_data = self.data.model_dump(
            mode="json", exclude_unset=True
        )

        # Compute dynamic task data once to extract subject and
        # configuration_context
        # TODO: these values can be computed in a separate methods if the rest
        # TODO: of compute_dynamic_data becomes too expensive to run twice
        # TODO: this may be replaced with using tags computed as system tags
        task_db = TaskDatabase(self.work_request)
        early_dynamic_task_data = self.compute_dynamic_data(task_db)

        if config_collection is not None:
            apply_configuration(
                self.work_request.configured_task_data,
                config_collection,
                self.TASK_TYPE,
                self.name,
                early_dynamic_task_data.subject,
                early_dynamic_task_data.configuration_context,
            )

        # Build the set of task-provided tags
        tagset = tags.make_task_provided_tagset()
        tagset.add(ttags.PROVENANCE_SYSTEM, self.compute_system_provided_tags())
        tagset.add(
            ttags.PROVENANCE_USER,
            self.compute_user_provided_tags(early_dynamic_task_data),
        )
        tagset.finalize()
        self.work_request.scheduler_tags_provided = sorted(tagset.tags)
        del tagset

        # Build the set of task-required tags
        tagset = tags.make_task_required_tagset()
        tagset.add(wtags.PROVENANCE_SYSTEM, self.compute_system_required_tags())
        tagset.finalize()
        self.work_request.scheduler_tags_required = sorted(tagset.tags)


class DefaultDynamicData[TD: BaseTaskData](
    DBTask[TD, BaseDynamicTaskData], metaclass=ABCMeta
):
    """Base class for tasks that do not add to dynamic task data."""

    @override
    def build_dynamic_data(
        self,
        task_database: TaskDatabaseInterface,  # noqa: U100
    ) -> BaseDynamicTaskData:
        """Return default dynamic data."""
        return BaseDynamicTaskData()


class DBWorkerTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    DBTask[TD, DTD], metaclass=ABCMeta
):
    """Base class for DB tasks of type WORKER."""

    TASK_TYPE = TaskTypes.WORKER

    @override
    def _execute(self) -> WorkRequestResults:  # noqa: C901
        raise AssertionError("Worker tasks cannot be executed server-side")


class DBProxyTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    DBTask[TD, DTD], metaclass=ABCMeta
):
    """A :py:class:`DBTask` that wraps a non-db task."""

    #: Wrapped BaseTask subclass
    task_class: type[BaseTask[TD, DTD]]
    #: Wrapped BaseTask
    task: BaseTask[TD, DTD]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Disable automatic registration."""
        # Do not register proxies in the type system automatically: it is done
        # by populate_task_registry

    @classmethod
    def populate_task_registry(cls) -> None:
        """Populate DBWorkerTask types from the BaseTask registry."""
        registry = DBTask._sub_tasks[cls.TASK_TYPE]
        for name, task_cls in BaseTask._sub_tasks[cls.TASK_TYPE].items():
            # Skip names already registered, so that the method can be run
            # twice to catch newly registered classes
            if name in registry:
                continue
            db_task_cls = type(
                f"DB{name}",
                (cls,),
                {
                    "TASK_VERSION": task_cls.TASK_VERSION,
                    "name": task_cls.name,
                    "task_data_type": task_cls.task_data_type,
                    "dynamic_task_data_type": task_cls.dynamic_task_data_type,
                    "task_class": task_cls,
                    "__doc__": task_cls.__doc__,
                },
            )
            registry[name] = db_task_cls

    def __init__(self, work_request: WorkRequest) -> None:
        """Initialize the task with the wrapped BaseTask."""
        super().__init__(work_request)
        self.task = self.task_class(
            task_data=self.work_request.used_task_data,
            dynamic_task_data=self.work_request.dynamic_task_data,
        )

    @override
    def build_architecture(self) -> str | None:
        return self.task.build_architecture()

    @override
    def compute_system_required_tags(self) -> set[str]:
        tags = super().compute_system_required_tags()
        tags |= self.task.compute_system_required_tags()
        return tags

    @override
    def build_dynamic_data(self, task_database: TaskDatabaseInterface) -> DTD:
        raise NotImplementedError(
            "build_dynamic_data cannot be called directly,"
            " only via compute_dynamic_data"
        )

    @override
    def compute_dynamic_data(self, task_database: TaskDatabaseInterface) -> DTD:
        task_database.resolve_inputs(self.task)
        return self.task.build_dynamic_data()

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        if not super().can_run_on(worker_metadata):
            return False
        return self.task.can_run_on(worker_metadata)

    @override
    def set_worker(self, worker: Worker) -> None:
        worker_metadata = worker.metadata()
        # TODO: Drop system:host_architecture fallback once all workers
        # have been updated.
        self.task.worker_native_architecture = worker_metadata.get(
            "system:native_architecture",
            worker_metadata.get("system:host_architecture"),
        )


class DBWorkerProxyTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    DBWorkerTask[TD, DTD], DBProxyTask[TD, DTD], metaclass=ABCMeta
):
    """A :py:class:`DBTask` that wraps a worker task."""

    @override
    @classmethod
    def populate_task_registry(cls) -> None:
        # Ensure worker tasks are registered before creating their proxy
        # classes
        import debusine.tasks  # noqa: F401

        super().populate_task_registry()


class DBSigningProxyTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    DBProxyTask[TD, DTD], metaclass=ABCMeta
):
    """A :py:class:`DBTask` that wraps a worker task."""

    TASK_TYPE = TaskTypes.SIGNING

    @override
    @classmethod
    def populate_task_registry(cls) -> None:
        # Ensure signing tasks are registered before creating their proxy
        # classes
        import debusine.signing.tasks  # noqa: F401

        super().populate_task_registry()

    @override
    def _execute(self) -> WorkRequestResults:  # noqa: C901
        raise AssertionError("Signing tasks cannot be executed server-side")


DBWorkerProxyTask.populate_task_registry()
DBSigningProxyTask.populate_task_registry()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the workspace UI helper."""

from django.test import RequestFactory

from debusine.db.models import Workspace
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.ui.workspaces import WorkspaceUI


class WorkspaceUITests(TestCase):
    """Test the :py:class:`WorkspaceUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def get_helper(self, workspace: Workspace) -> WorkspaceUI:
        """Instantiate a UI helper for a workspace."""
        factory = RequestFactory()
        request = factory.get("/")
        return WorkspaceUI(request, workspace)

    def test_candidate_parents_empty(self) -> None:
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [])

    def test_candidate_parents_other_scope(self) -> None:
        other_scope = self.playground.get_or_create_scope("other")
        other_workspace = self.playground.create_workspace(
            name="other", scope=other_scope, public=True
        )
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [other_workspace])

    def test_candidate_parents_only_visible(self) -> None:
        self.playground.create_workspace(name="other")
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [])

    def test_candidate_parents_current_parents(self) -> None:
        """Current parents are candidates even if not normally accessible."""
        public = self.playground.create_workspace(name="public", public=True)
        private = self.playground.create_workspace(name="private")
        self.scenario.workspace.set_inheritance([private, public])
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [private, public])

    def test_candidate_parents_ordering(self) -> None:
        other_scope = self.playground.get_or_create_scope("other")
        b = self.playground.create_workspace(name="b", public=True)
        a = self.playground.create_workspace(name="a", public=True)
        c = self.playground.create_workspace(name="c", public=True)
        ob = self.playground.create_workspace(
            name="b", public=True, scope=other_scope
        )
        oa = self.playground.create_workspace(
            name="a", public=True, scope=other_scope
        )
        oc = self.playground.create_workspace(
            name="c", public=True, scope=other_scope
        )
        helper = self.get_helper(self.scenario.workspace)
        self.assertEqual(helper.candidate_parents, [a, b, c, oa, ob, oc])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for task input fields."""

from typing import Any
from unittest import mock

from debusine.artifacts.models import CollectionCategory
from debusine.tasks.executors import ExecutorImageCategory, ExecutorInterface
from debusine.tasks.models import BackendType, LookupSingle
from debusine.tasks.tests.helper_mixin import FakeTaskDatabase
from debusine.test import TestCase


class TaskDatabaseInterfaceTests(TestCase):
    """Tests for :py:class:`TaskDatabaseInterface`."""

    def patch_executor(
        self, backend: BackendType, image_category: ExecutorImageCategory
    ) -> mock.MagicMock:
        """Patch executor for `backend` and return its mock."""
        executor = mock.MagicMock(spec=ExecutorInterface)
        executor.image_category = image_category
        self.enterContext(
            mock.patch.dict(
                "debusine.tasks.executors.base._backends", {backend: executor}
            )
        )
        return executor

    def assertLooksUp(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None,
        /,
        **kwargs: Any,
    ) -> None:
        task_db = FakeTaskDatabase()
        with mock.patch.object(
            task_db, "lookup_single_artifact", return_value="RETURN"
        ) as lookup_single_artifact:
            self.assertEqual(task_db.get_environment(**kwargs), "RETURN")

        lookup_single_artifact.assert_called_once_with(
            lookup, default_category=default_category
        )

    def test_get_environment_int(self) -> None:
        self.assertLooksUp(1, None, lookup=1)

    def test_get_environment_with_architecture_and_backend(self) -> None:
        self.assertLooksUp(
            "debian/match:codename=bookworm:format=tarball:backend=unshare",
            None,
            lookup="debian/match:codename=bookworm",
            backend=BackendType.UNSHARE,
        )

    def test_get_environment_with_architecture_from_worker(self) -> None:
        self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.IMAGE)
        self.assertLooksUp(
            "debian/match:codename=bookworm:architecture=amd64:"
            "format=image:backend=unshare:variant=",
            None,
            lookup="debian/match:codename=bookworm",
            architecture="amd64",
            backend=BackendType.UNSHARE,
            try_variant="",
        )

    def test_get_environment_with_architecture_missing(self) -> None:
        self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.IMAGE)
        self.assertLooksUp(
            "debian/match:codename=bookworm:format=image:"
            "backend=unshare:variant=",
            None,
            lookup="debian/match:codename=bookworm",
            backend=BackendType.UNSHARE,
            try_variant="",
        )

    def test_get_environment_tries_variant(self) -> None:
        self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.TARBALL)
        self.assertLooksUp(
            "debian/match:codename=bookworm:architecture=amd64:"
            "format=tarball:backend=unshare:"
            "variant=samplebasetaskwithexecutorandarchitecture",
            None,
            lookup="debian/match:codename=bookworm",
            architecture="amd64",
            backend=BackendType.UNSHARE,
            try_variant="samplebasetaskwithexecutorandarchitecture",
        )

    def test_get_environment_without_try_variant(self) -> None:
        self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.TARBALL)
        self.assertLooksUp(
            "debian/match:codename=bookworm:architecture=amd64:"
            "format=tarball:backend=unshare",
            None,
            lookup="debian/match:codename=bookworm",
            architecture="amd64",
            backend=BackendType.UNSHARE,
            try_variant=None,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db collections."""

import datetime as dt
import re
from collections.abc import Generator
from functools import cached_property
from typing import Any, ClassVar, Optional, Self, TYPE_CHECKING, TypeAlias

import jsonpath_rw
from django.conf import settings
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import RangeOperators
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import (
    CheckConstraint,
    Count,
    F,
    JSONField,
    Q,
    QuerySet,
    UniqueConstraint,
    Value,
)
from django.db.models.fields.json import KT
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    BaseArtifactDataModel,
    CollectionCategory,
    DebusinePromise,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.constraints import JsonDataUniqueConstraint
from debusine.db.context import context
from debusine.db.models.permissions import (
    Allow,
    PermissionUser,
    permission_check,
    permission_filter,
)
from debusine.db.models.workspaces import Workspace

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.artifacts import Artifact
    from debusine.db.models.auth import User
    from debusine.db.models.work_requests import WorkRequest
    from debusine.server.collections import CollectionManagerInterface
    from debusine.web.views.ui.collections import CollectionItemUI, CollectionUI
else:
    TypedModelMeta = object

#: Regexp matching the structure of collection names
collection_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_collection_name(value: str) -> bool:
    """Check if value is a valid scope name."""
    return value == "_" or bool(collection_name_regex.match(value))


def validate_collection_name(value: str) -> None:
    """Validate collection names."""
    if not is_valid_collection_name(value):
        raise ValidationError(
            "%(value)r is not a valid collection name", params={"value": value}
        )


class _CollectionRetainsArtifacts(models.TextChoices):
    """Choices for Collection.retains_artifacts."""

    NEVER = "never", "Never"
    WORKFLOW = "workflow", "While workflow is running"
    ALWAYS = "always", "Always"


class CollectionQuerySet[A](QuerySet["Collection", A]):
    """Custom QuerySet for Collection."""

    def in_current_scope(self) -> Self:
        """Filter to collections in the current scope."""
        return self.filter(workspace__scope=context.require_scope())

    def in_current_workspace(self) -> Self:
        """Filter to collections in the current workspace."""
        return self.filter(workspace=context.require_workspace())

    @permission_filter(workers=Allow.ALWAYS, anonymous=Allow.PASS)
    def can_display(self, user: PermissionUser) -> Self:
        """Keep only Collections that can be displayed."""
        # Delegate to workspace can_display check
        return self.filter(workspace__in=Workspace.objects.can_display(user))

    @permission_filter()
    def can_configure(self, user: PermissionUser) -> Self:
        """Keep only Collections that can be configured."""
        # Delegate to workspace can_configure check
        return self.filter(workspace__in=Workspace.objects.can_configure(user))

    @permission_filter()
    def can_delete(self, user: PermissionUser) -> Self:
        """Keep only Collections that can be deleted."""
        # Delegate to workspace can_configure check
        return self.filter(
            workspace__in=Workspace.objects.can_configure(user)
        ).exclude(category=CollectionCategory.WORKFLOW_INTERNAL)

    def exported_suites(self) -> "CollectionQuerySet[Any]":
        """Filter to exported ``debian:suite`` collections."""
        return (
            self.filter(category=CollectionCategory.SUITE)
            .annotate(
                exported=Coalesce(
                    "data__exported", Value("true"), output_field=JSONField()
                )
            )
            .exclude(exported=False)
        )


class CollectionManager(models.Manager["Collection"]):
    """Manager for Collection model."""

    def get_queryset(self) -> CollectionQuerySet[Any]:
        """Use the custom QuerySet."""
        return CollectionQuerySet(self.model, using=self._db)

    @staticmethod
    def get_or_create_singleton(
        category: CollectionCategory,
        workspace: Workspace,
        *,
        data: dict[str, Any] | None = None,
    ) -> "tuple[Collection, bool]":
        """Create a singleton collection."""
        if category not in SINGLETON_COLLECTION_CATEGORIES:
            raise ValueError(
                f"'{category}' is not a singleton collection category"
            )
        return Collection.objects.get_or_create(
            name="_", category=category, workspace=workspace, data=data or {}
        )


class Collection(models.Model):
    """Model representing a collection."""

    RetainsArtifacts: TypeAlias = _CollectionRetainsArtifacts

    name = models.CharField(
        max_length=255, validators=[validate_collection_name]
    )
    category = models.CharField(max_length=255)
    full_history_retention_period = models.DurationField(
        null=True,
        blank=True,
        help_text="Period after removing an item from the collection"
        " during which the related artifact is not allowed to expire",
    )
    metadata_only_retention_period = models.DurationField(
        null=True,
        blank=True,
        help_text="Period after removing an item from the collection during"
        " which its metadata is kept in the history of the collection",
    )

    workspace = models.ForeignKey(
        Workspace, on_delete=models.PROTECT, related_name="collections"
    )
    retains_artifacts = models.CharField(
        max_length=8,
        choices=RetainsArtifacts.choices,
        default=RetainsArtifacts.ALWAYS,
        help_text="Controls if artifacts in this collection should be retained"
        " rather than being allowed to expire",
    )

    child_artifacts = models.ManyToManyField(
        "db.Artifact",
        through="db.CollectionItem",
        through_fields=("parent_collection", "artifact"),
        related_name="parent_collections",
    )
    child_collections = models.ManyToManyField(
        "self",
        through="db.CollectionItem",
        through_fields=("parent_collection", "collection"),
        related_name="parent_collections",
        symmetrical=False,
    )

    data = models.JSONField(default=dict, blank=True)

    objects = CollectionManager.from_queryset(CollectionQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                fields=["name", "category", "workspace"],
                name="%(app_label)s_%(class)s_unique_name_category_workspace",
            ),
            CheckConstraint(
                check=~Q(name=""), name="%(app_label)s_%(class)s_name_not_empty"
            ),
            CheckConstraint(
                check=(
                    (
                        ~Q(category__in=SINGLETON_COLLECTION_CATEGORIES)
                        & ~Q(name__startswith="_")
                    )
                    | Q(category__in=SINGLETON_COLLECTION_CATEGORIES, name="_")
                ),
                name="%(app_label)s_%(class)s_name_not_reserved",
            ),
            CheckConstraint(
                check=~Q(category=""),
                name="%(app_label)s_%(class)s_category_not_empty",
            ),
        ]

    def __str__(self) -> str:
        """Return id, name, category."""
        # Stringify using a valid lookup syntax
        return f"{self.name}@{self.category}"

    def __repr__(self) -> str:
        """Return id, name, category."""
        # Stringify using a valid lookup syntax
        return f"<Collection: {self.name}@{self.category} ({self.pk})>"

    def get_absolute_url(self) -> str:
        """Return the canonical URL to display the collection."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:collections:detail",
                kwargs={
                    "wname": self.workspace.name,
                    "ccat": self.category,
                    "cname": self.name,
                },
            )

    def get_absolute_url_search(self) -> str:
        """Return the canonical URL to search inside the collection."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:collections:search",
                kwargs={
                    "wname": self.workspace.name,
                    "ccat": self.category,
                    "cname": self.name,
                },
            )

    @cached_property
    def manager(self) -> "CollectionManagerInterface":
        """Get collection manager for this collection category."""
        # Local import to avoid circular dependency
        from debusine.server.collections import CollectionManagerInterface

        return CollectionManagerInterface.get_manager_for(self)

    def ui(self, request: "HttpRequest") -> "CollectionUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.collections import CollectionUI

        return CollectionUI.get(request, self)

    @permission_check(
        "{user} cannot display collection {resource}",
        workers=Allow.ALWAYS,
        anonymous=Allow.PASS,
    )
    def can_display(self, user: PermissionUser) -> bool:
        """Check if the collection can be displayed."""
        return self.workspace.can_display(user)

    @permission_check(
        "{user} cannot configure collection {resource}",
    )
    def can_configure(self, user: PermissionUser) -> bool:
        """Check if the collection can be configured."""
        return self.workspace.can_configure(user)

    @permission_check(
        "{user} cannot delete collection {resource}",
    )
    def can_delete(self, user: PermissionUser) -> bool:
        """Check if the collection can be deleted."""
        return (
            self.category != CollectionCategory.WORKFLOW_INTERNAL
            and self.workspace.can_configure(user)
        )

    def get_item_stats(
        self,
    ) -> Generator[dict[str, Any]]:
        """Return a dict of item counts for each item type."""
        for rec in (
            self.child_items.values(
                "child_type", "category", active=Q(removed_at__isnull=True)
            )
            .annotate(count=Count("id"))
            .order_by("child_type", "category", "active")
        ):
            child_type = CollectionItem.Types(rec["child_type"])
            yield {
                "type": child_type,
                "category": rec["category"],
                "active": rec["active"],
                "count": rec["count"],
            }

    def delete_collection(self) -> None:
        """
        Delete this collection, and all its CollectionItem entries.

        Does not delete the associated Collections or Artifacts.
        """
        self.child_items.all().delete()
        self.delete()


class CollectionItemQuerySet(QuerySet["CollectionItem"]):
    """Custom QuerySet for CollectionItem."""

    def active(self) -> "CollectionItemQuerySet":
        """Filter to only active collection items."""
        return super().filter(removed_at__isnull=True)

    def active_at(self, at: dt.datetime) -> "CollectionItemQuerySet":
        """Filter to collection items that were active at a given time."""
        return self.filter(
            Q(created_at__lte=at),
            Q(removed_at__isnull=True) | Q(removed_at__gt=at),
        )

    def artifacts_in_collection(
        self, parent_collection: Collection, category: ArtifactCategory
    ) -> "CollectionItemQuerySet":
        """
        Filter to artifacts in a given parent collection.

        This is tuned to use the indexes on :py:class:`CollectionItem`.
        """
        return self.filter(
            parent_collection=parent_collection,
            # Technically redundant with parent_collection, but clues
            # PostgreSQL into using the correct index.
            parent_category=parent_collection.category,
            child_type=CollectionItem.Types.ARTIFACT,
            category=category,
        )


class CollectionItemManager(models.Manager["CollectionItem"]):
    """Manager for CollectionItem model."""

    @staticmethod
    def create_from_bare_data(
        category: BareDataCategory,
        *,
        parent_collection: Collection,
        name: str,
        data: BaseArtifactDataModel | dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"],
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from bare data."""
        if isinstance(data, BaseArtifactDataModel):
            data = data.model_dump(mode="json", exclude_unset=True)

        match category:
            case BareDataCategory.PROMISE:
                # Raise ValueError if data is not valid
                DebusinePromise(**data)

        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            child_type=CollectionItem.Types.BARE,
            category=category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    @staticmethod
    def create_from_artifact(
        artifact: "Artifact",
        *,
        parent_collection: Collection,
        name: str,
        data: dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"] = None,
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from the artifact."""
        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    @staticmethod
    def create_from_collection(
        collection: Collection,
        *,
        parent_collection: Collection,
        name: str,
        data: dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"] = None,
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from the collection."""
        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            collection=collection,
            child_type=CollectionItem.Types.COLLECTION,
            category=collection.category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    def drop_full_history(self, at: dt.datetime) -> None:
        """
        Drop artifacts from collections after full_history_retention_period.

        :param at: datetime to check if the artifacts are old enough.
        """
        self.get_queryset().exclude(removed_at__isnull=True).filter(
            # https://github.com/typeddjango/django-stubs/issues/1548
            removed_at__lt=(
                at - F("parent_collection__full_history_retention_period")  # type: ignore[operator] # noqa: E501
            )
        ).update(artifact=None)

    def drop_metadata(self, at: dt.datetime) -> None:
        """
        Delete old collection items.

        After full_history_retention_period + metadata_only_retention_period.

        :param at: datetime to check if the collection item is old enough.
        """
        self.get_queryset().exclude(removed_at__isnull=True).filter(
            # https://github.com/typeddjango/django-stubs/issues/1548
            removed_at__lt=(
                at  # type: ignore[operator]
                - F("parent_collection__full_history_retention_period")
                - F("parent_collection__metadata_only_retention_period")
            )
        ).delete()


class _CollectionItemTypes(models.TextChoices):
    """Choices for the CollectionItem.type."""

    BARE = "b", "Bare"
    ARTIFACT = "a", "Artifact"
    COLLECTION = "c", "Collection"


class CollectionItem(models.Model):
    """CollectionItem model."""

    name = models.CharField(max_length=255)

    Types: TypeAlias = _CollectionItemTypes

    child_type = models.CharField(max_length=1, choices=Types.choices)

    # category duplicates the category of the artifact or collection of this
    # item, so when the underlying artifact or collection is deleted the
    # category is retained
    category = models.CharField(max_length=255)

    parent_collection = models.ForeignKey(
        Collection,
        on_delete=models.PROTECT,
        related_name="child_items",
    )
    parent_category = models.CharField(max_length=255)

    collection = models.ForeignKey(
        Collection,
        on_delete=models.PROTECT,
        related_name="collection_items",
        null=True,
    )
    artifact = models.ForeignKey(
        "Artifact",
        on_delete=models.PROTECT,
        related_name="collection_items",
        null=True,
    )

    data = models.JSONField(default=dict, blank=True)

    created_at = models.DateTimeField(default=timezone.now)
    created_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_created_%(class)s",
    )
    created_by_workflow = models.ForeignKey(
        "WorkRequest",
        on_delete=models.SET_NULL,
        null=True,
        related_name="workflow_created_%(class)s",
    )

    removed_at = models.DateTimeField(blank=True, null=True)
    removed_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        null=True,
        related_name="user_removed_%(class)s",
    )
    removed_by_workflow = models.ForeignKey(
        "WorkRequest",
        on_delete=models.SET_NULL,
        null=True,
        related_name="workflow_removed_%(class)s",
    )

    objects = CollectionItemManager.from_queryset(CollectionItemQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            JsonDataUniqueConstraint(
                fields=[
                    "category",
                    "data->>'codename'",
                    "data->>'architecture'",
                    "data->>'variant'",
                    "data->>'backend'",
                    "parent_collection",
                ],
                condition=Q(
                    parent_category=CollectionCategory.ENVIRONMENTS,
                    removed_at__isnull=True,
                ),
                nulls_distinct=False,
                name="%(app_label)s_%(class)s_unique_debian_environments",
            ),
            UniqueConstraint(
                fields=["name", "parent_collection"],
                condition=Q(removed_at__isnull=True),
                name="%(app_label)s_%(class)s_unique_active_name",
            ),
            # Prevent direct way to add a collection to itself.
            # It is still possible to add loops of collections. The Manager
            # should avoid it
            CheckConstraint(
                check=~Q(collection=F("parent_collection")),
                name="%(app_label)s_%(class)s_distinct_parent_collection",
            ),
            CheckConstraint(
                name="%(app_label)s_%(class)s_childtype_removedat_consistent",
                check=(
                    Q(
                        child_type=_CollectionItemTypes.BARE,
                        collection__isnull=True,
                        artifact__isnull=True,
                    )
                    | (
                        Q(
                            child_type=_CollectionItemTypes.ARTIFACT,
                            collection__isnull=True,
                        )
                        & (
                            Q(artifact__isnull=False)
                            | Q(removed_at__isnull=False)
                        )
                    )
                    | (
                        Q(
                            child_type=_CollectionItemTypes.COLLECTION,
                            artifact__isnull=True,
                        )
                        & (
                            Q(collection__isnull=False)
                            | Q(removed_at__isnull=False)
                        )
                    )
                ),
            ),
        ]
        indexes = [
            models.Index("name", name="%(app_label)s_ci_name_idx"),
            models.Index(
                F("parent_collection"),
                KT("data__package"),
                KT("data__version"),
                name="%(app_label)s_ci_suite_source_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.SOURCE_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__srcpkg_name"),
                KT("data__srcpkg_version"),
                name="%(app_label)s_ci_suite_binary_source_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.BINARY_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__path"),
                name="%(app_label)s_ci_suite_index_path_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.REPOSITORY_INDEX,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__path"),
                name="%(app_label)s_ci_archive_index_path_idx",
                condition=Q(
                    parent_category=CollectionCategory.ARCHIVE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.REPOSITORY_INDEX,
                ),
            ),
            # Optimize generating suite indexes.
            models.Index(
                F("parent_collection"),
                KT("data__component"),
                F("created_at"),
                F("removed_at"),
                name="%(app_label)s_ci_suite_all_sources_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.SOURCE_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__component"),
                KT("data__architecture"),
                F("created_at"),
                F("removed_at"),
                name="%(app_label)s_ci_suite_all_binaries_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.BINARY_PACKAGE,
                ),
            ),
        ]

    def __str__(self) -> str:
        """Return id, name, collection_id, child_type."""
        item_info = (
            f" Artifact id: {self.artifact.id}"
            if self.artifact
            else (
                f" Collection id: {self.collection.id}"
                if self.collection
                else ""
            )
        )

        return (
            f"Id: {self.id} Name: {self.name} "
            f"Parent collection id: {self.parent_collection_id} "
            f"Child type: {self.child_type}"
            f"{item_info}"
        )

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Populate `parent_category` on save."""
        if not self.parent_category:
            self.parent_category = self.parent_collection.category
        super().save(*args, **kwargs)

    def get_absolute_url(self) -> str:
        """Return the canonical URL to display the item."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.parent_collection.workspace.scope.name):
            return reverse(
                "workspaces:collections:item_detail",
                kwargs={
                    "wname": self.parent_collection.workspace.name,
                    "ccat": self.parent_collection.category,
                    "cname": self.parent_collection.name,
                    "iid": self.pk,
                    "iname": self.name,
                },
            )

    @staticmethod
    def expand_variables(
        variables: dict[str, str], reference_data: dict[Any, Any]
    ) -> dict[str, str]:
        """
        Expand JSONPath variables against some data.

        The data will normally come from an Artifact.
        """
        jsonpaths = {}
        for name, path in variables.items():
            if name.startswith("$"):
                try:
                    jsonpaths[name[1:]] = jsonpath_rw.parse(path)
                except Exception as e:
                    raise ValueError(e)

        expanded_variables = {}
        for name, jsonpath in jsonpaths.items():
            matches = jsonpath.find(reference_data)
            if len(matches) == 1:
                expanded_variables[name] = matches[0].value
            elif len(matches) > 1:
                raise ValueError(
                    "Too many values expanding",
                    variables[f"${name}"],
                    reference_data,
                )
            else:
                raise KeyError(variables[f"${name}"], reference_data)

        for name, value in variables.items():
            if not name.startswith("$"):
                if name in expanded_variables:
                    raise ValueError(
                        f"Cannot set both '${name}' and '{name}' variables"
                    )
                expanded_variables[name] = value

        return expanded_variables

    @staticmethod
    def expand_name(
        name_template: str, expanded_variables: dict[str, str]
    ) -> str:
        """Format item name following item_template."""
        return name_template.format(**expanded_variables)

    def ui(self, request: "HttpRequest") -> "CollectionItemUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.collections import CollectionItemUI

        return CollectionItemUI.get(request, self)


class CollectionItemMatchConstraint(models.Model):
    """
    Enforce matching-value constraints on collection items.

    All instances of this model with the same :py:attr:`collection`,
    :py:attr:`constraint_name`, and :py:attr:`key` must have the same
    :py:attr:`value`.
    """

    objects: ClassVar[models.Manager["CollectionItemMatchConstraint"]] = (
        models.Manager["CollectionItemMatchConstraint"]()
    )

    collection = models.ForeignKey(
        Collection,
        on_delete=models.CASCADE,
        related_name="item_match_constraints",
    )
    # This is deliberately a bare ID, not a foreign key: some constraints
    # take into account even items that no longer exist but were in a
    # collection in the past.  Also note that the collection item's parent
    # collection may not be the same as this model's collection field: some
    # collections impose indirect constraints via child collections.
    collection_item_id = models.BigIntegerField()
    constraint_name = models.CharField(max_length=255)
    key = models.TextField()
    value = models.TextField()

    class Meta(TypedModelMeta):
        constraints = [
            ExclusionConstraint(
                name="%(app_label)s_%(class)s_match_value",
                expressions=(
                    (F("collection"), RangeOperators.EQUAL),
                    (F("constraint_name"), RangeOperators.EQUAL),
                    (F("key"), RangeOperators.EQUAL),
                    (F("value"), RangeOperators.NOT_EQUAL),
                ),
            )
        ]
        indexes = [
            models.Index(
                name="%(app_label)s_cimc_collection_item_idx",
                fields=["collection_item_id"],
            )
        ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Schedule WorkRequests to Workers."""

import logging
from operator import attrgetter
from typing import Any
from uuid import uuid4

import django.db
from celery import shared_task, states
from django.conf import settings
from django.db import connection, transaction
from django.db.models import (
    DateTimeField,
    Exists,
    F,
    IntegerField,
    JSONField,
    OuterRef,
    Q,
    Window,
)
from django.db.models.fields.json import KT
from django.db.models.functions import Cast, RowNumber
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django_celery_results.models import TaskResult

import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest, Worker
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.celery import run_server_task
from debusine.server.workflows.celery import run_workflow_task
from debusine.tasks import TaskConfigError
from debusine.tasks.models import Confirmation, OutputData, OutputDataError

logger = logging.getLogger(__name__)


class AssignToWorkerError(Exception):
    """Assigning a work request to a worker failed."""

    def __init__(self, message: str, code: str) -> None:
        """Construct the exception."""
        self.message = message
        self.code = code


def _possible_to_schedule_for_worker(worker: Worker) -> bool:
    if worker.is_at_capacity():
        # Worker already has enough work requests assigned; no scheduling
        # needed
        return False

    try:
        Worker.objects.select_for_update(nowait=True).get(id=worker.id)
    except django.db.DatabaseError:
        logger.debug("[SCHEDULER] Failed to lock Worker %s", worker)
        return False

    return True


@receiver(post_save, sender=Worker, dispatch_uid="ScheduleForWorker")
def _worker_changed(
    sender: type[Worker],  # noqa: ARG001, U100
    instance: Worker,
    **kwargs: Any,
) -> None:
    if not getattr(settings, "DISABLE_AUTOMATIC_SCHEDULING", False):
        schedule_for_worker(instance)


@transaction.atomic
def _assign_work_request_to_worker(
    work_request: WorkRequest, worker: Worker
) -> WorkRequest | None:
    """
    Attempt to assign a particular work request to the worker.

    :param work_request: the work request to try to assign.
    :param worker: the worker that needs a new work request to be
      assigned; must already be locked for update.
    :return: The assigned work request, or None if it could not be assigned.
    :raises AssignToWorkerError: if the work request is broken.
    """
    # Disable tag-based scheduling by default for now, but allow to enable it
    # with a feature flag so we can test it, before offering it for mainstream
    # adoption
    FEATURE_TAG_BASED_SCHEDULING = getattr(
        settings, "FEATURE_TAG_BASED_SCHEDULING", False
    )

    try:
        task = work_request.get_task(worker=worker)
    except TaskConfigError as exc:
        raise AssignToWorkerError(
            f"Failed to configure: {exc}", "configure-failed"
        ) from exc

    if not FEATURE_TAG_BASED_SCHEDULING and not task.can_run_on(
        worker.metadata()
    ):
        return None

    try:
        # We don't need to try to lock the whole workflow here, since
        # WorkRequest.assign_worker has no cascading effects on other work
        # requests.
        work_request = WorkRequest.objects.select_for_update(nowait=True).get(
            id=work_request.id
        )
    except django.db.DatabaseError:
        logger.debug(
            "[SCHEDULER] Failed to lock WorkRequest %s",
            work_request,
        )
        transaction.set_rollback(True)
        return None

    if work_request.worker:  # pragma: no cover
        # work_request did not have a worker assigned on the
        # initial qs but has one now - nothing to do
        return None

    with work_request.scheduling_disabled():
        try:
            work_request.assign_worker(worker)
        except SkipWorkRequest:
            logger.debug(
                "Skipping work request %s due to on_assignment action",
                work_request,
            )
            return None
        except Exception as exc:
            raise AssignToWorkerError(
                f"Failed to compute dynamic data: {exc}", "dynamic-data-failed"
            ) from exc

    if task.TASK_TYPE is TaskTypes.SERVER:
        # Use a slightly more informative task ID than the default.
        task_id = f"{work_request.task_name}_{work_request.id}_{uuid4()}"
        transaction.on_commit(
            lambda: run_server_task.apply_async(
                args=(work_request.id,), task_id=task_id
            )
        )

    return work_request


def _assign_work_request_to_worker_or_error(
    work_request: WorkRequest, worker: Worker
) -> WorkRequest | None:
    """
    Attempt to assign a particular work request to the worker, handling errors.

    This needs special care around transaction management.  Assigning the
    worker shouldn't take a whole-workflow lock, because it doesn't need to
    and doing so would increase contention; but if the work request is
    broken then we must take a whole-workflow lock because marking the work
    request as completed might cascade to other work requests in the same
    workflow, and that must therefore be done in a separate savepoint from
    the assignment to avoid a potential deadlock.

    :param work_request: the work request to try to assign.
    :param worker: the worker that needs a new work request to be
      assigned; must already be locked for update.
    :return: The assigned work request, or None if it could not be assigned.
    """
    try:
        return _assign_work_request_to_worker(work_request, worker)
    except AssignToWorkerError as exc:
        with transaction.atomic():
            # _assign_work_request_to_worker runs in its own savepoint, so
            # any update locks it holds will be released in case an
            # exception is raised, allowing this to be called in the
            # exception handler without risking a deadlock on WorkRequest
            # locks.  We must still use nowait=True, both because we're in
            # the scheduler and can't block, and because we're within a
            # Worker update lock and don't want to risk something else
            # trying to acquire WorkRequest and Worker locks in the opposite
            # order.
            try:
                WorkRequest.objects.filter(
                    id=work_request.id
                ).lock_workflows_for_update(nowait=True)
            except django.db.DatabaseError:
                # Not ideal, but we'll try the same work request again
                # later.
                logger.debug(
                    "[SCHEDULER] Failed to lock WorkRequest %s to record error",
                    work_request,
                )
                transaction.set_rollback(True)
            else:
                logger.error(
                    "Error assigning work request %s/%s (%s): %s",
                    work_request.task_type,
                    work_request.task_name,
                    work_request.id,
                    exc.message,
                )
                work_request.mark_completed(
                    WorkRequest.Results.ERROR,
                    output_data=OutputData(
                        errors=[
                            OutputDataError(message=exc.message, code=exc.code)
                        ]
                    ),
                )
        return None


def get_provided_worker_tags(worker: Worker) -> set[str]:
    """Get the set of tags provided by a worker."""
    from debusine.server.tags import make_worker_provided_tagset

    tagset = make_worker_provided_tagset()
    tagset.add(
        wtags.PROVENANCE_WORKER, worker.dynamic_metadata.get("provided_tags")
    )
    tagset.add(
        wtags.PROVENANCE_ADMIN, worker.static_metadata.get("provided_tags")
    )
    tagset.finalize()
    return tagset.tags


def get_required_task_tags(worker: Worker) -> set[str]:
    """Get the set of task tags required by a worker."""
    from debusine.server.tags import make_worker_required_tagset

    tagset = make_worker_required_tagset()
    tagset.add(
        wtags.PROVENANCE_ADMIN, worker.static_metadata.get("required_tags")
    )
    tagset.finalize()
    return tagset.tags


@transaction.atomic
def schedule_for_worker(worker: Worker) -> WorkRequest | None:
    """
    Schedule a new work request for the worker.

    :param worker: the worker that needs a new work request to be
      assigned.
    :return: The assigned work request. None if no suitable work request could
      be found.
    :rtype: WorkRequest | None.
    """
    # Disable tag-based scheduling by default for now, but allow to enable it
    # with a feature flag so we can test it, before offering it for mainstream
    # adoption
    FEATURE_TAG_BASED_SCHEDULING = getattr(
        settings, "FEATURE_TAG_BASED_SCHEDULING", False
    )

    if not _possible_to_schedule_for_worker(worker):
        return None

    provided_worker_tags = get_provided_worker_tags(worker)
    required_task_tags = get_required_task_tags(worker)
    worker_metadata = worker.metadata()
    tasks_allowlist = worker_metadata.get("tasks_allowlist", None)
    tasks_denylist = worker_metadata.get("tasks_denylist", [])

    # TODO: Optimization: pre-filter on task type as appropriate.
    qs = WorkRequest.objects.pending(exclude_assigned=True).filter(
        task_type__in=(
            TaskTypes.WORKER,
            TaskTypes.SERVER,
            TaskTypes.SIGNING,
        )
    )
    if tasks_allowlist is not None:
        qs = qs.filter(task_name__in=tasks_allowlist)
    elif tasks_denylist:
        qs = qs.exclude(task_name__in=tasks_denylist)

    if FEATURE_TAG_BASED_SCHEDULING:
        qs = qs.filter(
            scheduler_tags_required__contained_by=list(provided_worker_tags),
            scheduler_tags_provided__contains=list(required_task_tags),
        )

    for work_request in qs:
        assigned_work_request = _assign_work_request_to_worker_or_error(
            work_request, worker
        )
        if assigned_work_request is not None:
            return assigned_work_request

    return None


def run_workflow_task_via_celery(work_request: WorkRequest) -> None:
    """Run a workflow callback or workflow via Celery."""
    # Use a slightly more informative task ID than the default.
    task_id = (
        f"{work_request.task_type.lower()}_{work_request.task_name}_"
        f"{work_request.id}_{uuid4()}"
    )
    transaction.on_commit(
        lambda: run_workflow_task.apply_async(
            args=(work_request.id,), task_id=task_id
        )
    )


@transaction.atomic
def schedule_internal() -> list[WorkRequest]:
    """
    Schedule internal work requests.

    These don't require a worker, but need some kind of action from the
    scheduler.
    """
    in_progress_task_results = TaskResult.objects.filter(
        status__in={
            states.PENDING,
            states.RECEIVED,
            states.STARTED,
            states.RETRY,
        },
        task_name=("debusine.server.workflows.celery.run_workflow_task"),
    ).annotate(
        task_args_array=Cast("task_args", JSONField()),
        first_task_arg=Cast("task_args_array__0", IntegerField()),
    )

    pending_wrs = WorkRequest.objects.pending()
    pending_synchronization_points = pending_wrs.filter(
        task_type=TaskTypes.INTERNAL, task_name="synchronization_point"
    )
    pending_workflow_callbacks = (
        pending_wrs.filter(task_type=TaskTypes.INTERNAL, task_name="workflow")
        .annotate(
            row_number=Window(
                RowNumber(), partition_by=F("parent"), order_by="id"
            )
        )
        # Only one workflow callback for a given workflow may run at once.
        .exclude(
            Exists(
                in_progress_task_results.filter(
                    first_task_arg__in=WorkRequest.objects.filter(
                        parent=OuterRef(OuterRef("parent"))
                    ),
                )
            )
        )
        .exclude(row_number__gt=1)
    )
    pending_workflows = (
        pending_wrs.filter(task_type=TaskTypes.WORKFLOW)
        .annotate(
            row_number=Window(
                RowNumber(), partition_by=F("workflow_root"), order_by="id"
            )
        )
        # Only one sub-workflow within a given root workflow may run at once.
        .exclude(
            Exists(
                in_progress_task_results.filter(
                    first_task_arg=OuterRef("workflow_root")
                )
            )
        )
        .exclude(row_number__gt=1)
    )
    pending_waits = pending_wrs.filter(task_type=TaskTypes.WAIT)
    pending_combined = WorkRequest.objects.filter(
        Q(id__in=pending_synchronization_points)
        | Q(id__in=pending_workflow_callbacks)
        | Q(id__in=pending_workflows)
        | Q(id__in=pending_waits)
    )

    try:
        pending_combined.lock_workflows_for_update(nowait=True)
    except django.db.DatabaseError:
        # Something like select_for_update(skip_locked=True) might be
        # better, allowing us to make some progress, but it's unclear how to
        # acquire a lock on only those work requests that are in workflows
        # that can be locked as a whole without blocking.  This should be
        # fine provided that nothing holds a lock for too long on workflows
        # that had pending internal work requests at the start of its
        # transaction.
        logger.debug(
            "[SCHEDULER] Failed to lock workflows with pending internal"
            " work requests"
        )
        return []

    workflow_roots: set[WorkRequest] = set()
    result: list[WorkRequest] = []
    for work_request in pending_combined:
        match (work_request.task_type, work_request.task_name):
            case (TaskTypes.INTERNAL, "synchronization_point"):
                # Synchronization points do nothing; when pending, they are
                # immediately marked as completed, thus unblocking work
                # requests that depend on them.
                work_request.mark_completed(WorkRequest.Results.SUCCESS)
                result.append(work_request)

            case (TaskTypes.INTERNAL, "workflow"):
                # Workflow callbacks are handled by the corresponding
                # workflow orchestrator.  This may take a while, so we
                # delegate it to a Celery task.
                work_request.mark_running()
                run_workflow_task_via_celery(work_request)
                result.append(work_request)

            case (TaskTypes.WORKFLOW, _):
                # Workflow population always happens via the orchestrator
                # for the root workflow.  Sub-workflow orchestrators are not
                # called directly from the scheduler, although if a
                # sub-workflow is pending then its root workflow may deal
                # with populating it.
                # Workflows always have a root.
                assert work_request.workflow_root is not None
                workflow_roots.add(work_request.workflow_root)

            case (TaskTypes.WAIT, _):
                # Wait tasks are marked running immediately, but otherwise
                # we do nothing; some later event will mark them as
                # completed.
                work_request.mark_running()
                result.append(work_request)

            case _ as unreachable:
                raise AssertionError(
                    f"Unexpected internal task name: {unreachable}"
                )

    for workflow_root in sorted(workflow_roots, key=attrgetter("id")):
        # Workflows are populated by the corresponding workflow
        # orchestrator.  This may take a while, so we delegate it to a
        # Celery task.
        workflow_root.mark_running()
        run_workflow_task_via_celery(workflow_root)
        result.append(workflow_root)

    return result


@transaction.atomic
def complete_wait_tasks() -> None:
    """Mark certain wait tasks as completed."""
    qs = (
        WorkRequest.objects.running()
        .filter(task_type=TaskTypes.WAIT)
        .filter(
            # Delay tasks complete automatically once their delay has
            # expired.
            Q(task_name="delay")
            # Confirm tasks with auto_confirm_if_no_failure=True complete
            # automatically if all their dependencies are successful.
            | (
                Q(
                    task_name="confirm",
                    task_data__auto_confirm_if_no_failure=True,
                )
                & ~Exists(
                    WorkRequest.objects.filter(
                        Q(reverse_dependencies=OuterRef("id"))
                        & ~Q(
                            status=WorkRequest.Statuses.COMPLETED,
                            result=WorkRequest.Results.SUCCESS,
                        )
                    )
                )
            )
        )
        .annotate(
            delay_until=Cast(
                KT("task_data__delay_until"),
                DateTimeField(blank=True, null=True),
            )
        )
        .exclude(task_name="delay", delay_until__gt=timezone.now())
    )

    try:
        qs.lock_workflows_for_update(nowait=True)
    except django.db.DatabaseError:
        # Something like select_for_update(skip_locked=True) might be
        # better, allowing us to make some progress, but it's unclear how to
        # acquire a lock on only those work requests that are in workflows
        # that can be locked as a whole without blocking.  This should be
        # fine provided that nothing holds a lock for too long on workflows
        # that had pending internal work requests at the start of its
        # transaction.
        logger.debug(
            "[SCHEDULER] Failed to lock workflows with pending internal"
            " work requests"
        )
        return

    for work_request in qs:
        match work_request.task_name:
            case "delay":
                work_request.mark_completed(WorkRequest.Results.SUCCESS)
            case "confirm":
                work_request.mark_completed(
                    WorkRequest.Results.SUCCESS,
                    output_data=OutputData(
                        confirmation=Confirmation(confirmed=True)
                    ),
                )
            case _ as unreachable:
                raise AssertionError(f"Unexpected task name: {unreachable}")


@receiver(post_save, sender=WorkRequest, dispatch_uid="ScheduleForWorkRequest")
def _work_request_changed(
    sender: type[WorkRequest],  # noqa: ARG001, U100
    instance: WorkRequest,
    **kwargs: Any,
) -> None:
    if not getattr(
        settings, "DISABLE_AUTOMATIC_SCHEDULING", False
    ) and not hasattr(instance, "_disable_signals"):
        # Run the scheduler on commit, but only once per savepoint.  This
        # uses undocumented Django internals, but if it goes wrong without
        # an obvious unit test failure, then the worst case should be some
        # extra calls to the scheduler, which is relatively harmless.
        if (set(connection.savepoint_ids), schedule_task.delay) not in [
            hook[:2] for hook in connection.run_on_commit
        ]:
            transaction.on_commit(schedule_task.delay)


def schedule() -> list[WorkRequest]:
    """
    Try to assign work requests to free workers.

    The function loops over workers that have no work request assigned
    and tries to assign them new work requests.

    :return: the list of work requests that got assigned.
    """
    available_workers = Worker.objects.waiting_for_work_request()

    result = list(schedule_internal())
    for available_worker in available_workers:
        work_request = schedule_for_worker(available_worker)
        if work_request:
            result.append(work_request)

    complete_wait_tasks()

    return result


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task()  # type: ignore[untyped-decorator]
def schedule_task() -> None:
    """Run the scheduler as a celery task."""
    schedule()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Input artifacts for tasks."""

import abc
from collections.abc import Collection as ABCCollection
from typing import Any, TYPE_CHECKING, assert_never, override

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import RelationType
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.models import (
    ExtraDebusineRepository,
    ExtraExternalRepository,
    ExtraRepository,
    LookupMultiple,
    LookupSingle,
)
from debusine.tasks.server import (
    ArtifactInfo,
    CollectionInfo,
    MultipleArtifactInfo,
    TaskDatabaseInterface,
)

if TYPE_CHECKING:
    from debusine.tasks import BaseTask


class TaskInput[RD](abc.ABC):
    """Representation for one or more input artifacts for a task."""

    name: str

    @abc.abstractmethod
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> RD:
        """Resolve the input from task data."""

    def __set_name__(
        self, owner: type["BaseTask[Any, Any]"], name: str
    ) -> None:
        """Set the field name at class construction."""
        self.name = name
        # Avoid sharing the inputs definition across all subclasses of BaseTask
        if "inputs" not in owner.__dict__:
            owner.inputs = owner.inputs.copy()
        owner.inputs[name] = self

    def __get__(
        self,
        obj: "BaseTask[Any, Any]",
        _: type["BaseTask[Any, Any]"] | None = None,
    ) -> RD:
        """Look up this input in the task object."""
        from debusine.tasks import TaskConfigError

        # If __get__ is called, it means that self.name is not in obj.__dict__,
        # which only happens if resolve_to_task has not yet been called
        raise TaskConfigError(f"Cannot access unresolved input {self.name!r}")

    def resolve_to_task(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> None:
        """Resolve this input and set the result in the task dict."""
        task.__dict__[self.name] = self.resolve(task, task_database)

    @staticmethod
    def resolve_inputs(
        task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> None:
        """Resolve all input fields in the given task."""
        for inp in task.inputs.values():
            inp.resolve_to_task(task, task_database)


class DataFieldInput[RD](TaskInput[RD]):
    """Input field that resolves values based on a task data field."""

    field: str

    def __init__(self, *, field: str | None = None) -> None:
        """
        Declare an input defined by a task data field.

        :param field: dot-separated list of lookup names in task data. Defaults
          to the input member name
        """
        if field is not None:
            self.field = field

    def __set_name__(
        self, owner: type["BaseTask[Any, Any]"], name: str
    ) -> None:
        """Set the field name at class construction."""
        super().__set_name__(owner, name)
        if not hasattr(self, "field"):
            self.field = name

    def _resolve_field(self, task_data: Any) -> Any:
        """Look up the value corresponding to self.field in task_data."""
        # Prevent import loop
        from debusine.tasks import TaskConfigError

        try:
            lookup: Any = task_data
            for name in self.field.split("."):
                lookup = getattr(lookup, name)
        except AttributeError:
            raise TaskConfigError(f"Invalid input field: {self.field!r}")
        return lookup

    def _get_lookup_single(
        self,
        task_data: Any,
    ) -> LookupSingle | None:
        """Find the LookupSingle identified by self.field in task_data."""
        lookup = self._resolve_field(task_data)
        assert lookup is None or isinstance(lookup, LookupSingle)
        return lookup

    def _get_lookup_multiple(self, task_data: Any) -> LookupMultiple:
        """Find the LookupMultiple identified by self.field in task_data."""
        lookup = self._resolve_field(task_data)
        assert isinstance(lookup, LookupMultiple)
        return lookup

    def _get_lookup_list(
        self,
        task_data: Any,
    ) -> list[LookupSingle | LookupMultiple]:
        """Find the list of lookups identified by self.field in task_data."""
        lookups = self._resolve_field(task_data)
        if lookups is None:
            return []
        assert isinstance(lookups, list)
        for lookup in lookups:
            assert isinstance(lookup, LookupSingle | LookupMultiple)
        return lookups


class ArtifactInput[RD](DataFieldInput[RD]):
    """Look up artifacts enforcing their types."""

    def __init__(
        self,
        *,
        field: str | None = None,
        categories: ABCCollection[ArtifactCategory] | None = None,
    ) -> None:
        """
        Look up artifacts enforcing their type.

        :param field: dot-separated list of lookup names in task data
        :param categories: list of acceptable artifact categories
        """
        super().__init__(field=field)
        self.categories: tuple[ArtifactCategory, ...] | None = (
            tuple(categories) if categories is not None else None
        )

    def check_category(
        self, info: ArtifactInfo, field_name: str | None = None
    ) -> None:
        """Check the category of an ArtifactInfo."""
        # Prevent import loop
        from debusine.tasks import ensure_artifact_categories

        if self.categories is None:
            return

        ensure_artifact_categories(
            configuration_key=field_name or self.field,
            category=info.category,
            expected=self.categories,
        )


class SingleInput(ArtifactInput[ArtifactInfo]):
    """One single input artifact specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> ArtifactInfo:
        lookup = self._get_lookup_single(task.data)
        assert lookup is not None
        info = task_database.lookup_single_artifact(lookup)
        self.check_category(info)
        return info


class SingleInputList(ArtifactInput[list[ArtifactInfo]]):
    """A list of single input artifacts specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> list[ArtifactInfo]:
        infos: list[ArtifactInfo] = []
        for lookup in self._get_lookup_list(task.data):
            assert isinstance(lookup, LookupSingle)
            info = task_database.lookup_single_artifact(lookup)
            self.check_category(info)
            infos.append(info)
        return infos


class OptionalSingleInput(ArtifactInput[ArtifactInfo | None]):
    """One optional single input artifact specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> ArtifactInfo | None:
        lookup = self._get_lookup_single(task.data)
        info = task_database.lookup_single_artifact(lookup)
        if info is not None:
            self.check_category(info)
        return info


class MultiInput(ArtifactInput[MultipleArtifactInfo]):
    """A list of multiple artifacts lookups specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> MultipleArtifactInfo:
        lookup = self._get_lookup_multiple(task.data)
        assert lookup is not None
        info = task_database.lookup_multiple_artifacts(lookup)
        if self.categories is not None:
            for i, artifact in enumerate(info):
                self.check_category(artifact, f"{self.field}[{i}]")
        return info


class MultiInputList(ArtifactInput[list[MultipleArtifactInfo]]):
    """A multiple artifacts lookup specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> list[MultipleArtifactInfo]:
        infos: list[MultipleArtifactInfo] = []
        for lookup_idx, lookup in enumerate(self._get_lookup_list(task.data)):
            assert isinstance(lookup, LookupMultiple)
            info = task_database.lookup_multiple_artifacts(lookup)
            if self.categories is not None:
                for result_idx, artifact in enumerate(info):
                    self.check_category(
                        artifact, f"{self.field}[{lookup_idx}][{result_idx}]"
                    )
            infos.append(info)
        return infos


class UploadArtifactsInput(ArtifactInput[MultipleArtifactInfo]):
    """A multiple artifact lookup finding all related binaries in an upload."""

    @override
    def __init__(
        self,
        *,
        field: str | None = None,
        categories: ABCCollection[ArtifactCategory] | None = None,
    ) -> None:
        categories_with_upload = [ArtifactCategory.UPLOAD]
        categories_with_upload.extend(categories or ())
        super().__init__(field=field, categories=categories_with_upload)

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> MultipleArtifactInfo:
        lookup = self._get_lookup_multiple(task.data)
        assert lookup is not None

        # Incrementally build the result set
        result: list[ArtifactInfo] = []

        # Lookup the initial set of packages
        infos = task_database.lookup_multiple_artifacts(lookup)
        upload_ids: list[int] = []
        for i, info in enumerate(infos):
            if info.category == ArtifactCategory.UPLOAD:
                # File uploads for later expansion
                upload_ids.append(info.id)
            else:
                self.check_category(info, f"{self.field}[{i}]")
                result.append(info)

        # Expand with binaries related to uploads
        if upload_ids:
            result.extend(
                task_database.find_related_artifacts(
                    upload_ids,
                    ArtifactCategory.BINARY_PACKAGE,
                    relation_type=RelationType.EXTENDS,
                )
            )

        return MultipleArtifactInfo(result)


class EnvironmentInput(ArtifactInput[ArtifactInfo]):
    """One environment artifact specified in a task data field."""

    def __init__(
        self,
        *,
        field: str | None = None,
        categories: ABCCollection[ArtifactCategory] | None = None,
        image_category: ExecutorImageCategory | None = None,
        set_backend: bool = True,
        try_variant: bool = True,
    ) -> None:
        """
        Look up artifacts enforcing their type.

        :param field: dot-separated list of lookup names in task data
        :param categories: list of acceptable artifact categories
        :param image_category: try to use an environment with this image
          category; defaults to the image category needed by the executor
          for `self.backend`
        :param set_backend: if True (default), try to use an environment
          matching `self.backend`
        :param try_variant: if True (default), try to use an environment
          whose variant is `self.name`, but fall back to looking up an
          environment without a variant if the first lookup fails
        """
        super().__init__(field=field, categories=categories)
        self.image_category = image_category
        self.set_backend = set_backend
        self.try_variant = try_variant

    @override
    def resolve(
        self, task: "BaseTask[Any,Any]", task_database: TaskDatabaseInterface
    ) -> ArtifactInfo:
        # Prevent import loop
        from debusine.tasks import BaseTaskWithExecutor

        assert isinstance(task, BaseTaskWithExecutor)
        lookup = self._get_lookup_single(task.data)
        assert lookup is not None
        info = task_database.get_environment(
            lookup,
            architecture=task.build_architecture(),
            backend=task.backend if self.set_backend else None,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=self.image_category,
            try_variant=task.name if self.try_variant else None,
        )
        self.check_category(info)
        return info


class OptionalEnvironmentInput(ArtifactInput[ArtifactInfo | None]):
    """One optional environment artifact specified in a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any,Any]", task_database: TaskDatabaseInterface
    ) -> ArtifactInfo | None:
        # Prevent import loop
        from debusine.tasks import BaseTaskWithExecutor

        assert isinstance(task, BaseTaskWithExecutor)
        lookup = self._get_lookup_single(task.data)
        if lookup is None:
            return None
        info = task_database.get_environment(
            lookup,
            architecture=task.build_architecture(),
            backend=task.backend,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant=task.name,
        )
        self.check_category(info)
        return info


class SuiteArchiveInput(
    DataFieldInput[tuple[CollectionInfo, CollectionInfo | None]]
):
    """Look up a suite and an archive based on a task data field."""

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> tuple[CollectionInfo, CollectionInfo | None]:
        # Prevent import loop
        from debusine.tasks import ensure_collection_category

        lookup = self._get_lookup_single(task.data)
        assert lookup is not None

        suite = task_database.lookup_single_collection(lookup)
        ensure_collection_category(
            configuration_key=self.field,
            category=suite.category,
            expected=CollectionCategory.SUITE,
        )

        archive = task_database.lookup_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        return suite, archive


class ExtraRepositoriesInput(
    DataFieldInput[list[ExtraExternalRepository] | None]
):
    """Look up extra repositories."""

    def _resolve_extra_repository(
        self,
        task_database: TaskDatabaseInterface,
        extra_repository: ExtraRepository,
    ) -> ExtraExternalRepository:
        """Resolve a single extra repository into the external form."""
        # Prevent import loop
        from debusine.tasks import TaskConfigError

        match extra_repository:
            case ExtraDebusineRepository():
                debusine_fqdn = task_database.get_server_setting(
                    "DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN"
                )
                suite = task_database.lookup_single_collection(
                    extra_repository.suite
                )
                components = extra_repository.components
                if components is None:
                    if "components" not in suite.data:
                        raise TaskConfigError(
                            f"'components' not set for {extra_repository.suite}"
                        )
                    components = suite.data["components"]
                return ExtraExternalRepository(
                    # Ideally we'd use HTTPS, but that might require
                    # certificate setup within environments, and we know that
                    # our own repositories are signed, so HTTP should be good
                    # enough to use from workers for now.
                    url=(
                        f"http://{debusine_fqdn}"
                        f"/{suite.scope_name}/{suite.workspace_name}"
                    ),
                    suite=suite.name,
                    components=components,
                    signing_key=task_database.export_suite_signing_keys(
                        suite.id
                    ),
                )
            case ExtraExternalRepository():
                return extra_repository
            case _ as unreachable:
                assert_never(unreachable)

    @override
    def resolve(
        self, task: "BaseTask[Any, Any]", task_database: TaskDatabaseInterface
    ) -> list[ExtraExternalRepository] | None:
        # Prevent import loop
        from debusine.tasks import ExtraRepositoryMixin

        assert isinstance(task, ExtraRepositoryMixin)

        extra_repositories = self._resolve_field(task.data)
        if extra_repositories is None:
            return None
        return [
            self._resolve_extra_repository(task_database, extra_repository)
            for extra_repository in extra_repositories
        ]


class DebusineFQDNInput(TaskInput[str]):
    """Resolve to the Debusine FQDN."""

    @override
    def resolve(
        self, task: "BaseTask[Any,Any]", task_database: TaskDatabaseInterface
    ) -> str:
        return task_database.get_server_setting("DEBUSINE_FQDN")
